"""
Get SC Labs Test Result Data
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/8/2022
Updated: 9/5/2022
License: MIT License <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Tools to collect SC Labs test results from CoA PDFs and URLs.

Data Points:

    ✓ analyses
    ✓ {analysis}_method
    ✓ {analysis}_status
    ✓ coa_urls
    ✓ date_collected
    ✓ date_tested
    ✓ date_received
    ✓ distributor
    ✓ distributor_address
    ✓ distributor_street
    - distributor_city
    - distributor_state
    ✓ distributor_zipcode
    ✓ distributor_license_number
    ✓ images
    ✓ lab_results_url
    ✓ producer
    ✓ producer_address
    - producer_street
    ✓ producer_city
    - producer_state
    - producer_zipcode
    ✓ producer_license_number
    ✓ product_name
    ✓ lab_id
    ✓ product_type
    ✓ batch_number
    ✓ metrc_ids
    - metrc_lab_id
    ✓ metrc_source_id
    - product_size
    - serving_size
    - servings_per_package
    - sample_weight
    ✓ results
    ✓ status
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_cbg
    ✓ total_thcv
    ✓ total_cbc
    ✓ total_cbdv
    ✓ total_terpenes
    ✓ sample_id (generated)
    - strain_name (augmented)
    ✓ lab
    ✓ lab_image_url
    - lab_license_number
    ✓ lab_address
    ✓ lab_street
    ✓ lab_city
    ✓ lab_county (augmented)
    ✓ lab_state
    ✓ lab_zipcode
    ✓ lab_phone
    - lab_email
    ✓ lab_website
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)

Data Sources:
    
    - SC Labs Test Results
    URL: <https://client.sclabs.com/>

Future work:

    - Optional: Normalize the `results`, `images`.
    - Optional: Standardize the `product_type` and `status`.
    - Optional: Find the `county` given the `producer_zipcode`.
    - Future work: Identify a `strain_name` from the `product_name`.

"""
# Internal imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from time import sleep
from typing import Any
from urllib.parse import urljoin

# External imports.
from bs4 import BeautifulSoup
import pandas as pd
import pdfplumber
import requests

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
    parse_data_block,
)
from cannlytics.utils.constants import (
    ANALYSES,
    ANALYTES,
    DEFAULT_HEADERS,
    STANDARD_FIELDS,
    STANDARD_UNITS,
)
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    strip_whitespace,
)


# It is assumed that the lab has the following details.
SC_LABS = {
    'coa_algorithm': 'sclabs.py',
    'coa_algorithm_entry_point': 'parse_sc_labs_coa',
    'lims': 'SC Labs',
    'url': 'https://client.sclabs.com',
    'lab': 'SC Labs',
    'lab_image_url': 'https://www.sclabs.com/wp-content/uploads/2020/11/sc-labs-logo-white.png',
    # 'lab_license_number': '', # Get this data point dynamically.
    'lab_address': '100 Pioneer St., Ste. E, Santa Cruz, CA 95060',
    'lab_street': '100 Pioneer St., Ste. E',
    'lab_city': 'Santa Cruz',
    'lab_county': 'Santa Cruz',
    'lab_state': 'CA',
    'lab_zipcode': '95060',
    'lab_phone': '(866) 435-0709',
    'lab_email': 'info@sclabs.com',
    'lab_website': 'https://sclabs.com',
    'lab_latitude': 36.987869,
    'lab_longitude': -122.033162,
}

# It is assumed that the CoA has the following parameters.
SC_LABS_COA = {
    'coa_distributor_area': '(205, 150, 400, 230)',
    'coa_producer_area': '(0, 150, 204.0, 230)',
    'coa_page_area': [
        '(0, 80, 305, 720)',
        '(305, 80, 612, 720)',
    ],
    'coa_sample_details_area': [
        '(0, 225, 200, 350)',
        '(200, 225, 400, 350)',
    ],
}


def get_sc_labs_test_results(
        producer_id,
        reverse=True,
        limit=100,
        page_limit=100,
        pause=0.2,
        headers=None
    ) -> list:
    """Get all test results for a specific SC Labs client.
    Args:
        producer_id (str): A producer ID.
        reverse (bool): Whether to collect in reverse order, True by default (optional).
        limit (int): The number of samples per page to collect, 100 by default.
        page_limit (int): The maximum number of pages to collect, 100 by default.
        pause (float): A respectful pause to wait between requests.
    Returns:
        (list): A list of dictionaries of sample data.
    """

    #  Iterate over pages, getting all the samples on each page,
    # until the active page is repeated and the first sample is the same.
    active_page = None
    first_sample = None
    samples = []
    sample_pages = range(1, page_limit)
    if headers is None:
            headers = DEFAULT_HEADERS
    for sample_page in sample_pages:

        # Pause between requests to be respectful of the API server.
        if sample_page > 1 and pause:
            sleep(pause)

        # Get a client page with X amount of samples.
        url = '/'.join([SC_LABS['url'], 'client', str(producer_id)])
        params = {'limit': limit, 'page': sample_page}
        response = requests.get(url, headers=headers, params=params)
        soup = BeautifulSoup(response.content, 'html.parser')

        # Check if the page is a 404.
        top_spans = soup.find_all('span')
        for span in top_spans:
            if '404' in span.text:
                print('Client not found: %s' % (producer_id))
                break

        # Note: This may not be working as intended.
        # Break the iteration if the page max is reached.
        try:
            current_sample = soup.find('h3').text
        except AttributeError:
            current_sample = first_sample
        attributes = {'class': 'pagination-active'}
        current_page = soup.find('li', attrs=attributes)
        if (current_page == active_page) and (current_sample == first_sample):
            break
        active_page = current_page
        first_sample = current_sample

        # Get producer.
        details = soup.find('div', attrs={'id': 'detailQuickView'})
        try:
            producer = details.find('h2').text
        except AttributeError:
            try:
                producer = soup.find('h2').text
            except AttributeError:
                producer = 'Anonymous'

        # Get producer image.
        try:
            producer_image_url = details.find('img')['src'] \
                .replace('\n', '').strip()
        except AttributeError:
            producer_image_url = ''

        # Get producer website.
        try:
            attributes = {'class': 'pp-social-web'}
            el = details.find('span', attrs=attributes)
            producer_url = el.find('a')['href']
        except:
            producer_url = ''

        # Get all of the sample cards.
        cards = soup.find_all('div', attrs={'class': 'grid-item'})
        if reverse:
                cards.reverse()
        for card in cards:

            # Get the lab's internal ID.
            lab_id = card['id'].replace('div_', '')

            # Get the product name.
            product_name = card.find('h3').text

            # Get lab results URL.
            base_url = SC_LABS['url']
            href = card.find('a')['href']
            lab_results_url = urljoin(base_url, href)

            # Get the date tested.
            mm, dd, yyyy = card.find('h6').text.split('-')
            date = '-'.join([yyyy, mm, dd])

            # Get totals.
            totals = card.find('div', attrs={'class': 'sample-details'})
            values = totals.find_all('div')
            total_thc = values[0].text.split(':')[-1].replace('%', '')
            total_cbd = values[1].text.split(':')[-1].replace('%', '')
            total_terpenes = values[2].text.split(':')[-1].replace('%', '')
            
            # # Create a sample ID.
            # # Note: This may not work as intended if any lab results
            # # do not have tested at dates.
            # sample_id = create_sample_id(
            #     private_key=json.dumps(results),
            #     public_key=product_name,
            #     salt=producer,
            # )

            # Aggregate sample data.
            sample = {
                'date_received': date,
                'lab_id': lab_id,
                'lab_results_url': lab_results_url,
                'producer_id': producer_id,
                'producer': producer,
                'producer_image_url': producer_image_url,
                'product_name': product_name,
                'producer_url': producer_url,
                'sample_id': None,
                'total_cbd': total_cbd,
                'total_thc': total_thc,
                'total_terpenes': total_terpenes,
            }
            samples.append(sample)

    # Return all of the samples for the client.
    return samples


def get_sc_labs_client_details():
    """Find client details. This is useful when getting a sample's
    data directly.
    Returns:
        (dict): The client's details data.
    """
    # TODO: Implement.
    raise NotImplementedError


def get_sc_labs_sample_details(
        sample='',
        headers=None,
        **kwargs,
    ) -> dict:
    """Get the details for a specific SC Labs test sample.
    Args:
        sample (str): A sample number or sample URL.
        headers (dict): Headers for the HTTP request (optional).
    Returns:
        (dict): A dictionary of sample details.
    """

    # Get the sample page.
    base_url = SC_LABS['url']
    if sample.startswith(base_url):
        url = sample
    else:
        if headers is None:
            headers = DEFAULT_HEADERS
        url = urljoin(base_url, f'sample/{sample}/')
    response = requests.get(url, headers=headers)
    soup = BeautifulSoup(response.content, 'html.parser')

    # Get the bulk of the details.
    els = soup.find_all('p', attrs={'class': 'sdp-summary-data'})
    obs = parse_data_block(els)

    # Get the product name.
    obs['product_name'] = soup.find('h2').text

    # Format the distributor address.
    try:
        address = obs.get('address', '').split('*')[-1].strip()
        obs['distributor_address'] = address
        obs['distributor_city'] = address.split(',')[0]
        obs['distributor_zipcode'] = address.split(' ')[-1]
    except TypeError:
        obs['distributor_address'] = ''
        obs['distributor_city'] = ''
        obs['distributor_zipcode'] = ''
    
    # Get the producer details.
    el = soup.find('div', attrs={'id': 'cultivator-details'})
    producer_details = parse_data_block(el)
    obs['producer'] = producer_details['business_name']
    obs['producer_license_number'] = producer_details['license_number']

    # Format the producer address.
    try:
        address = producer_details['address'].split('*')[-1].strip()
        obs['producer_address'] = address
        obs['producer_city'] = address.split(',')[0]
        obs['producer_zipcode'] = address.split(' ')[-1]
    except TypeError:
        obs['producer_address'] = ''
        obs['producer_city'] = ''
        obs['producer_zipcode'] = ''
    
    # TODO: Lookup producer / distributor address for latitude / longitude.
    
    # Remove the `address` field to avoid confusion.
    try:
        del obs['address']
    except KeyError:
        pass

    # Get the producer image.
    el = soup.find('div', attrs={'id': 'clientprofileimage'})
    obs['producer_image_url'] = strip_whitespace(el.find('img')['src'])

    # Get the producer URL.
    links = soup.find_all('a', attrs={'class': 'greybutton'})
    for link in links:
        href = link['href']
        if 'sample' in href:
            sample_number = href.split('sample/')[-1].split('/')[0]
            obs['sample_number'] = sample_number
            obs['coa_urls'] = [{
                'url': urljoin(url, href),
                'filename': f'{sample_number}.pdf',
            }]
        elif 'client' in href:
            obs['producer_url'] = urljoin(url, href)

    # Get the Metrc IDs.
    try:
        metrc_ids = obs['source_metrc_uid'].split(',')
        obs['metrc_ids'] = [x.strip() for x in metrc_ids]
    except KeyError:
        obs['metrc_ids'] = []

    # Get the product type.
    try:
        attributes = {'class': 'sdp-producttype'}
        obs['product_type'] = soup.find('p', attrs=attributes).text
    except AttributeError:
        obs['product_type'] = 'Unknown'

    # Get the image.
    try:
        attributes = {'data-popup': 'fancybox'}
        image_url = soup.find('a', attrs=attributes)['href']
        obs['images'] = [{
            'url': image_url,
            'filename': image_url.split('/')[-1],
        }]
    except TypeError:
        obs['images'] = []

    # Get the date tested.
    try:
        el = soup.find('div', attrs={'class': 'sdp-masthead-data'})
        mm, dd, yyyy = el.find('p').text.split('/')
        obs['date_tested'] = '-'.join([yyyy, mm, dd])
    except AttributeError:
        obs['date_tested'] = ''

    # Get the overall status: Pass / Fail.
    try:
        status = soup.find('p', attrs={'class': 'sdp-result-pass'}).text
        obs['status'] = status.replace('\n', '').strip()
    except AttributeError:
        obs['status'] = ''

    # Format the dates.
    try:
        mm, dd, yyyy = obs['date_collected'].split('/')
        obs['date_collected'] = '-'.join([yyyy, mm, dd]) 
    except KeyError:
        obs['date_collected'] = ''
    try:
        mm, dd, yyyy = obs['date_received'].split('/')
        obs['date_received'] = '-'.join([yyyy, mm, dd])
    except KeyError:
        obs['date_received'] = ''
    
    # Rename desired fields.
    # Note: There may be a better way to do this.
    rename = {}
    for key, value in obs.items():
        try:
            standard_field = STANDARD_FIELDS[key]
            rename[standard_field] = value
        except KeyError:
            rename[key] = value
    obs = rename

    # Get the CoA ID.
    try:
        attributes = {'class': 'coa-id'}
        obs['coa_id'] = soup.find('p', attrs=attributes).text \
            .split(':')[-1]
    except AttributeError:
        obs['coa_id'] = ''

    # Remove any keys that begin with a digit.
    for key in list(obs.keys()):
        if not key or key[0].isdigit():
            del obs[key]

    # Optional: Try to get sample_weight.

    # Get all of the analyses and results.
    analyses = []
    results = []
    notes = None
    cards = soup.find_all('div', attrs={'class': 'analysis-container'})    
    for card in cards:

        # Get the analysis.
        analysis = card.find('h4').text
        if 'Notes' in analysis:
            div = card.find('div', attrs={'class': 'section-inner'})
            notes = div.find('p').text
        if 'Analysis' not in analysis:
            continue
        analysis = snake_case(analysis.split(' Analysis')[0])
        analysis = ANALYSES.get(analysis, analysis)
        analyses.append(analysis)

        # Get the method for the analysis.
        bold = card.find('b')
        method = bold.parent.text.replace('Method: ', '')
        key = '_'.join([analysis, 'method'])
        obs[key] = method

        # Get all of the results for the analysis.
        # - value, units, margin_of_error, lod, loq
        table = card.find('table')
        rows = table.find_all('tr')
        for row in rows[1:]:
            cells = row.find_all('td')
            result = {}
            for cell in cells:
                key = cell['class'][0].replace('table-', '')
                key = STANDARD_FIELDS.get(key, key)
                value = cell.text.replace('\n', '').strip()
                result[key] = value
                result['analysis'] = analysis
            results.append(result)
    
    # Parse legacy results if no results collected at this stage.
    if not results:

        # Get details.
        el = soup.find('div', attrs={'id': 'detailQuickView'})
        items = el.find_all('li')
        mm, dd, yyyy = items[0].text.split(': ')[-1].split('-')
        obs['date_tested'] = f'{yyyy}-{mm}-{dd}'
        obs['product_type'] = items[-1].text.split(': ')[-1]

        # Get analysis cards.
        cards = soup.find_all('div', attrs={'class': 'detail-row'})
        for card in cards:
            
            # Get the analysis.
            title = card.find('h3').text.lower()
            if 'not tested' in title:
                continue
            text = title.split('test')[0]
            analysis = snake_case(strip_whitespace(text))
            if analysis == 'label_claims':
                continue
            analysis = SC_LABS_COA['analyses'].get(analysis, analysis)

            # Get the method for the analysis.
            method = card.find('p').text
            key = '_'.join([analysis, 'method'])
            obs[key] = method
            
            # FIXME: This is not complete / throws an error.
            # Get all of the results for the analysis.
            # - value, units, margin_of_error, lod, loq
            table = card.find('table')
            rows = table.find_all('tr')
            for row in rows[1:]:
                cells = row.find_all('td')
                result = {}
                for cell in cells:
                    key = cell['class'][0].replace('table-', '')
                    key = STANDARD_FIELDS.get(key, key)
                    value = cell.text.replace('\n', '').strip()
                    result[key] = value
                    result['analysis'] = analysis
                results.append(result)

    # Separate `lod` and `loq`.
    lod_loq_values = [x['lodloq'].split(' / ') if x.get('lodloq') else None for x in results]
    for i, values in enumerate(lod_loq_values):
        if values is not None:
            result = results[i]
            result['lod'] = values[0]
            result['loq'] = values[1]
            del result['lodloq']
            results[i] = result

    # Clean results.
    for i, result in enumerate(results):

        # Clean the margin of error.
        try:
            margin = result['margin_of_error'].replace('±', '')
            result['margin_of_error'] = convert_to_numeric(margin)
        except KeyError:
            pass

        # Parse `units` from `value`. E.g. '71.742%'.
        value = result.get('value', '')
        result['value'] = re.sub('[^\d\.]', '', value)
        result['units'] = re.sub('[\d\.]', '', value)
        result['units'] = result['units'].replace('%', 'percent')

        # Try to ensure that the result values are numbers.
        result['value'] = convert_to_numeric(result['value'])
        result['mg_g'] = convert_to_numeric(result.get('mg_g'))
        result['lod'] = convert_to_numeric(result.get('lod'))
        result['loq'] = convert_to_numeric(result.get('loq'))
        result['limit'] = convert_to_numeric(result.get('limit'))

        # Assign a `key` for the analyte.
        analyte = snake_case(result['name'])
        result['key'] = parser.analytes(analyte, analyte)

        # Update the result.
        results[i] = result

    # Clean `total_{analyte}`s and `sum_of_cannabinoids`.
    columns = [x for x in obs.keys() if x.startswith('total_') or x.startswith('sum_')]
    for key in columns:
        obs[key] = convert_to_numeric(obs[key], strip=True)

    # Lowercase `{analysis}_status`
    columns = [x for x in obs.keys() if x.endswith('_status')]
    for key in columns:
        obs[key] = obs[key].lower()

    # Separate `batch_units` from `batch_size`.
    obs['batch_size'], obs['batch_units'] = tuple(obs['batch_size'].split(' '))

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Return the sample details with a new or re-minted `sample_id`.
    if not results:
        results = None
    obs['analyses'] = analyses
    obs['lab_results_url'] = url
    obs['notes'] = notes
    obs['results'] = results
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs['producer'],
    )
    obs['coa_parsed_at'] = datetime.now().isoformat()
    return { **SC_LABS, **obs}


def parse_sc_labs_pdf(parser, doc: Any, **kwargs) -> dict:
    """Parse a SC Labs CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    # Read the PDF.
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
    else:
        report = doc

    # Get the lab-specific CoA page areas, standard analyses and analytes.
    obs = {}
    coa_parameters = SC_LABS_COA
    distributor_area = literal_eval(coa_parameters['coa_distributor_area'])
    producer_area = literal_eval(coa_parameters['coa_producer_area'])
    sample_details_area = coa_parameters['coa_sample_details_area']

    # Get producer details.
    front_page = report.pages[0]
    crop = front_page.within_bbox(producer_area)
    details = crop.extract_text().replace('\n', '')
    address = details.split('Address:')[-1].strip()
    business = details.split('Business Name:')[-1].split('License Number:')[0].strip()
    license_number = details.split('License Number:')[-1].split('Address:')[0].strip()
    parts = address.split(',')
    street = parts[0]
    subparts = parts[-1].strip().split(' ')
    city = ' '.join(subparts[:-2])
    try:
        state, zipcode = subparts[-2], subparts[-1]
    except IndexError:
        state, zipcode = '', ''
    obs['producer'] = business
    obs['producer_address'] = address
    obs['producer_street'] = street
    obs['producer_city'] = city
    obs['producer_state'] = state
    obs['producer_zipcode'] = zipcode
    obs['producer_license_number'] = license_number

    # Get distributor details.
    crop = front_page.within_bbox(distributor_area)
    details = crop.extract_text().replace('\n', ' ')
    address = details.split('Address:')[-1].strip()
    business = details.split('Business Name:')[-1].split('License Number:')[0].strip()
    license_number = details.split('License Number:')[-1].split('Address:')[0].strip()
    parts = address.split(',')
    street = parts[0]
    subparts = parts[-1].strip().split(' ')
    city = ' '.join(subparts[:-2])
    try:
        state, zipcode = subparts[-2], subparts[-1]
    except IndexError:
        state, zipcode = '', ''
    obs['distributor'] = business
    obs['distributor_address'] = address
    obs['distributor_street'] = street
    obs['distributor_city'] = city
    obs['distributor_state'] = state
    obs['distributor_zipcode'] = zipcode
    obs['distributor_license_number'] = license_number

    # Get sample details.
    # FIXME: May be mishandling `sum_of_cannabinoids` and `total_cannabinoids`.
    if isinstance(sample_details_area, str):
        sample_details_area = [sample_details_area]
    for area in sample_details_area:
        crop = front_page.within_bbox(literal_eval(area))
        details = crop.extract_text().split('\n')
        for d in details:
            if ':' not in d:
                continue
            values = d.split(':')
            key = snake_case(values[0])
            key = STANDARD_FIELDS.get(key, key)
            obs[key] = values[-1]

    # Get the date tested, product name, and sample type.
    front_page_text = front_page.extract_text()
    date_tested = front_page_text.split('DATE ISSUED')[-1].split('|')[0].strip()
    lines = front_page_text.split('SAMPLE NAME:')[1].split('\n')
    product_name = lines[0].strip()
    obs['product_type'] = lines[1]

    # Get the analyses.
    analyses = []
    for i, line in enumerate(lines):
        if 'ANALYSIS' in line:
            analysis = line.split(' ANALYSIS')[0].lower()
            analysis = ANALYSES.get(analysis, analysis)
            if analysis == 'safety':
                parts = ' '.join(lines[i+1:i+3]).split(':')
                parts = [x.replace('PASS', '').replace('FAIL', '').strip() for x in parts]
                parts = [ANALYSES.get(x, snake_case(x)) for x in parts if x]
                analyses.extend(parts)
            else:
                analyses.append(analysis)

    # Get the cannabinoid and terpene totals.

    value = front_page_text.split('Sum of Cannabinoids:')[-1].split('%')[0]
    obs['sum_of_cannabinoids'] = convert_to_numeric(value, strip=True)

    value = front_page_text.split('Total Cannabinoids:')[-1].split('%')[0]
    obs['total_cannabinoids'] = convert_to_numeric(value, strip=True)

    value = front_page_text.split('Total THC:')[-1].split('%')[0]
    obs['total_thc'] = convert_to_numeric(value, strip=True)

    value = front_page_text.split('Total CBD:')[-1].split('%')[0]
    obs['total_cbd'] = convert_to_numeric(value, strip=True)

    value = front_page_text.split('Total Terpenoids:')[-1].split('%')[0]
    obs['total_terpenes'] = convert_to_numeric(value, strip=True)

    # Get the moisture content analysis if present.
    try:
        value = front_page_text.split('Moisture:')[-1].split('%')[0]
        obs['moisture_content'] = convert_to_numeric(value, strip=True)
        analyses.append('moisture')
    except:
        pass

    # Get all page text, from the 2nd page on, column by column.
    areas = coa_parameters['coa_page_area']
    lines = []
    for page in report.pages[1:]:
        for area in areas:
            crop = page.within_bbox(literal_eval(area))
            lines += crop.extract_text().split('\n')

    # Map all the analytes to analyses.
    # TODO: Is it possible to either make these field dynamic or
    # add these fields to the constants?
    analyte_analysis_map = {
        'Total Sample Area Covered by Sand, Soil, Cinders, or Dirt': 'foreign_matter',
        'Caryophyllene Oxide': 'terpenes',
        'Pentachloronitro- benzene*': 'pesticides',
        'Piperonylbu- toxide': 'pesticides',
        'DDVP (Dichlorvos)': 'pesticides',
        'Methyl parathion': 'pesticides',
        'Chlorantranilip- role': 'pesticides',
        'Clofentezine': 'pesticides',
        'Total Sample Area Covered by Sand, Soil, Cinders, or Dirt': 'foreign_matter',
        'Total Sample Area Covered by Mold': 'foreign_matter',
        'Total Sample Area Covered by an Imbedded Foreign Material': 'foreign_matter',
        'Insect Fragment Count': 'foreign_matter',
        'Hair Count': 'foreign_matter',
        'Mammalian Excreta Count': 'foreign_matter',
        'Shiga toxin-producing Escherichia coli': 'microbes',
        'Salmonella spp.': 'microbes',
        'Aspergillus fumigatus': 'microbes',
        'Aspergillus flavus': 'microbes',
        'Aspergillus niger': 'microbes',
        'Aspergillus terreus': 'microbes',
    }
    for line in lines:
        if 'TEST RESULT' in line:
            analysis_name = line.split('TEST RESULT')[0].strip().title()
            analysis = ANALYSES.get(analysis_name)
        elif 'Method:' in line:
            # FIXME: Imperfect method collect (missing text on next line).
            obs[f'{analysis}_method'] = line.split('Method:')[-1].strip()
        else:
            first_value = find_first_value(line)
            name = line[:first_value].replace('\n', ' ').strip()
            analyte_analysis_map[name] = analysis

    # Get the results.
    results = []
    for page in report.pages[1:]:

        # Get the results from each result page.
        tables = page.extract_tables()
        for table in tables:
            for i, row in enumerate(table):

                # Determine the analysis, then the units.
                analyte = row[0].replace('\n', ' ').strip()
                analysis = analyte_analysis_map.get(analyte)
                key = snake_case(analyte)
                key = ANALYTES.get(key, key)
                units = STANDARD_UNITS.get(analysis)

                # Skip per unit values.
                if 'per_unit' in key:
                    continue

                # Hot-fix: Skip (non-)analytes that begin with a digit.
                # Note: It would be best to improve this logic.
                if key[0].isdigit():
                    continue

                # Determine the values, handling screens differently.
                parts = row[1].split(' / ')
                subparts = parts[-1].strip().split(' ')
                limit = None
                status = None
                if len(subparts) == 3:
                    limit = subparts[1]
                    value, status = tuple(row[-1].split(' '))
                else:
                    try:
                        mg_g, value = tuple(row[-1].split(' '))
                    except ValueError:
                        value = row[-1]

                # Record the result
                results.append({
                    'analysis': analysis,
                    'key': key,
                    'limit': limit,
                    'lod': convert_to_numeric(parts[0].strip()),
                    'loq': convert_to_numeric(subparts[0]),
                    'margin_of_error': convert_to_numeric(subparts[-1].replace('±', '')),
                    'mg_g': convert_to_numeric(mg_g),
                    'name': analyte,
                    'status': status,
                    'units': units,
                    'value': convert_to_numeric(value),
                })

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    # TODO: Fix or make `sample_id` obsolete.
    obs = {**SC_LABS, **obs}
    obs['analyses'] = json.dumps(analyses)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['date_tested'] = date_tested
    obs['product_name'] = product_name
    obs['results'] = json.dumps(results)
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


def parse_sc_labs_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a TagLeaf LIMS CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    # Pythonic try first, ask later approach.
    # Optional: Be more precise and check if there is a URL and if it is private.
    try:
        data = get_sc_labs_sample_details(doc)
    except (AttributeError, ConnectionError):
        data = parse_sc_labs_pdf(parser, doc)
        data['public'] = False
        if isinstance(doc, str):
            data['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
        elif isinstance(doc, pdfplumber.pdf.PDF):
            data['coa_pdf'] = doc.stream.name.replace('\\', '/').split('/')[-1]
    return data


if __name__ == '__main__':

    # === Tests ===
    from cannlytics.data.coas import CoADoc

    # # [✓] TEST: Get all test results for a specific client.
    # test_results = get_sc_labs_test_results(sample='2821')
    # assert test_results is not None

    # # [✓] TEST: Get details for a specific sample ID.
    # sample_details = get_sc_labs_sample_details('858084')
    # assert sample_details is not None

    # [✓] TEST: Get details for a specific sample URL.
    # sc_labs_coa_url = 'https://client.sclabs.com/sample/858084'
    # parser = CoADoc()
    # lab = parser.identify_lims(sc_labs_coa_url)
    # assert lab == 'SC Labs'
    # data = get_sc_labs_sample_deta
    # ils(sc_labs_coa_url)
    # assert data is not None
    # print('Parsed:', doc)

    # [✓] TEST: Parse a SC Labs CoA PDF (with cannabinoids and terpenes).
    # directory = '../../../tests/assets/coas/sc-labs'
    # doc = f'{directory}/Mattole Valley Jack H.pdf'
    # parser = CoADoc()
    # lab = parser.identify_lims(doc)
    # assert lab == 'SC Labs'
    # data = parse_sc_labs_pdf(parser, doc)
    # assert data is not None
    # print('Parsed:', doc)

    # [✓] TEST: Parse a SC Labs CoA PDF (with safety screening).
    directory = '../../../tests/assets/coas/sc-labs'
    doc = f'{directory}/Cherry Punch.pdf'
    parser = CoADoc()
    lab = parser.identify_lims(doc)
    assert lab == 'SC Labs'
    data = parse_sc_labs_pdf(parser, doc)
    assert data is not None
    print('Parsed:', doc)

    # FIXME:
    # 220000981-Lime-Mojito-Joints.pdf
    # 220001004-Birthday-Cake-Joints.pdf
    # 220001026-Sweet-N-Sour-Joints.pdf
    # 220001027-Guava-Breeze-Joints.pdf
    # 220001033-Purple-Passion-Joints.pdf
    # 220001034-Unicorn-Cake-Joints.pdf
    # 220001035-Cr確e-de-Citron-Joints.pdf
    # 220001036-Kimbo-Cookie-Dough-Joints.pdf
    # 220001055-Cherry-Limeade-Joints.pdf
    # 220001061-Ice-Cream-Sandwich-Joints.pdf



"""
Parse CannaSafe CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 9/17/2022
Updated: 9/17/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse a CannaSafe CoA PDF.

Data Points:

"""
# Standard imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber

# Internal imports.
from cannlytics.data.data import create_sample_id
from cannlytics.utils.constants import STANDARD_UNITS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)

# It is assumed that the lab has the following details.
CANNASAFE = {
    'coa_algorithm': 'cannasafe.py',
    'coa_algorithm_entry_point': 'parse_cannasafe_coa',
    'lims': 'CannaSafe',
    'lab': 'CannaSafe',
    'lab_image_url': '',
    'lab_license_number': ' ', # <- Make dynamic.
    'lab_address': '', # <- Make dynamic.
    'lab_street': '', # <- Make dynamic.
    'lab_city': '', # <- Make dynamic.
    'lab_county': '', # <- Make dynamic.
    'lab_state': 'CA', # <- Make dynamic.
    'lab_zipcode': '', # <- Make dynamic.
    'lab_phone': '', # <- Make dynamic.
    'lab_email': '', # <- Make dynamic.
    'lab_website': '',
    'lab_latitude': 0, # <- Make dynamic.
    'lab_longitude': 0, # <- Make dynamic.
}


def parse_cannasafe_coa(
        parser,
        doc: Any,
        temp_path: Optional[str] = None,
        **kwargs,
    ) -> dict:
    """Parse a CannaSafe CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """

    # Get the lab's analyses.
    # lab_analyses = GREEN_LEAF_LAB_ANALYSES
    # coa_parameters = GREEN_LEAF_LAB_COA
    # standard_analyses = list(lab_analyses.keys())
    # analysis_names = [x['name'] for x in lab_analyses.values()]

    # TODO: If the `doc` is a URL, then download the PDF to `temp_path`.
    if temp_path is None:
        # FIXME: Get the user's temp_path
        temp_path = '/tmp'

    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]

    # Add optional ability to collect the image.
    # - Get the image data.
    # - Save the image to Firebase Storage
    # - Create a dynamic link for the image.

    # TODO: Get the lab details.

    # TODO: Get the sample details.

    # TODO: Get the client details.

    # TODO: Get the analyses.

    # TODO: Get the results.


# TEST:
url = 'https://tracker.csalabs.com/public/samples/lllora'
doc = '../../../tests/assets/coas/cannasafe/Grape-CBD-Gummies.pdf'

# Read the PDF.
obs = {}
report = pdfplumber.open(doc)
front_page = report.pages[0]


# TODO: Get the image.
# - Save the image locally from the image data.
# - Upload the image to Firebase Storage.
# - Create a dynamic link for the image (`sample_image_url`).
# img = front_page.images[3]

# Get the front page text.
text = front_page.extract_text()
# split = re.split('\d+ of \d+', text)
# header = split[0]
# body = split[-1].split('ND = Not Detected')[0]
# lines = body.split('\n')
# lines = [x for x in lines if x.strip()]



"""
Parse KCA Labs CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 9/17/2022
Updated: 9/19/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse KCA Labs CoA PDFs.

Data Points:

"""
# Standard imports.
from datetime import datetime
import json
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber
import requests

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import create_sample_id
from cannlytics.utils.constants import DEFAULT_HEADERS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)

# It is assumed that the lab has the following details.
KCA_LABS = {
    'coa_algorithm': 'kcalabs.py',
    'coa_algorithm_entry_point': 'parse_kcalabs_coa',
    'url': 'https://kcalabs.com',
    'lims': 'KCA Labs',
    'lab': 'KCA Labs',
    'lab_image_url': 'https://kcalabs.com/wp-content/uploads/2021/07/KCALABS_FI-2.png',
    'lab_license_number': 'P_0058',
    'lab_address': '232 North Plaza Drive, Nicholasville, KY 40356',
    'lab_street': '232 North Plaza Drive',
    'lab_city': 'Nicholasville',
    'lab_county': 'Jessamine',
    'lab_state': 'KY',
    'lab_zipcode': '40356',
    'lab_phone': '+1-833-KCA-LABS',
    'lab_email': 'trustedresults@kcalabs.com',
    'lab_website': 'https://kcalabs.com',
    'lab_latitude': 37.904740,
    'lab_longitude': -84.566980,
}
KCA_LABS_ANALYSES = {
    'cannabinoids': {
        'columns': ['name', 'lod', 'loq', 'value', 'mg_g'],
    },
    # 'pesticides': {
    #     'name': 'Pesticide Analysis by GCMS/LCMS',
    #     'columns': ['name', 'value', 'limit', 'lod', 'loq', 'units'],
    #     'double_column': True,
    # },
    # 'water_activity': {
    #     'name': 'Water Activity by Aqua Lab',
    #     'columns': ['name', 'value', 'units', 'lod', 'loq'],
    # },
    # 'moisture_content': {
    #     'name': 'Moisture by Moisture Balance',
    #     'columns': ['name', 'value', 'units'],
    # },
    # 'terpenes': {
    #     'name': 'Terpene Analysis by GCMS',
    #     'columns': ['name', 'value', 'mg_g'],
    #     'double_column': True,
    # },
    # 'heavy_metals': {
    #     'name': 'Metals Analysis by ICPMS',
    #     'columns': ['name', 'value', 'limit', 'lod', 'loq', 'units'],
    # },
    # 'mycotoxins': {
    #     'name': 'Mycotoxins by LCMSMS',
    #     'columns': ['name', 'value', 'limit', 'loq', 'loq', 'units'],
    # },
    # 'microbials': {
    #     'name': 'Microbials by PCR',
    #     'columns': ['name', 'status', 'limit', 'lod', 'loq', 'units', 'value'],
    # },
    # 'foreign_matter': {
    #     'name': 'Filth and Foreign Material Inspection by Magnification',
    #     'columns': ['name', 'status', 'limit', 'lod', 'loq', 'units'],
    # },
}
KCA_LABS_COA = {
   'coa_skip_values': [
        'Summary',
        'Test Date Tested Status',
        'Normalization',
        'LOD LOQ Result Result',
        'Analyte',
        '(%) (%) (%) (mg/unit)',
    ],
    'coa_sample_detail_fields': {
        'Sample ID': 'lab_id',
        'Batch': 'batch_number',
        'Received': 'received_at',
        'Type': 'product_type',
        'Completed': 'completed_at',
        'Matrix': 'product_subtype',
        'Unit Mass (g)': 'sample_weight',
    },
}

# Constants
coa_skip_values = [
    'Summary',
    'Test Date Tested Status',
    'Normalization',
    'LOD LOQ Result Result',
    'Analyte',
    '(%) (%) (%) (mg/unit)',
]
coa_sample_detail_fields = {
    'Sample ID': 'lab_id',
    'Batch': 'batch_number',
    'Received': 'date_received',
    'Type': 'product_type',
    'Completed': 'date_completed',
    'Matrix': 'product_subtype',
    'Unit Mass (g)': 'sample_weight',
}
coa_numeric_columns = [
    'sample_weight',
]


def parse_kcalabs_coa(
        parser,
        doc: Any,
        temp_path: Optional[str] = None,
        **kwargs,
    ) -> dict:
    """Parse a KCA Labs CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    obs = {}

    # TODO: Get the lab's parameters.
    # lab_analyses = GREEN_LEAF_LAB_ANALYSES
    # coa_parameters = GREEN_LEAF_LAB_COA
    # standard_analyses = list(lab_analyses.keys())
    # analysis_names = [x['name'] for x in lab_analyses.values()]

    # If the `doc` is a URL, then download the PDF to `temp_path`.
    # Otherwise open a PDF file or use a PDF instance.
    if isinstance(doc, str):
        if doc.startswith('https'):
            if temp_path is None:
                # FIXME: Get the user's temp_path
                temp_path = '/tmp'
            filename = doc.split('/')[-1]
            filepath = f'{temp_path}/{filename}.pdf'
            # FIXME: Will need to get `coa_url`.
            with open(filepath, 'wb') as f:
                f.write(requests.get(url, headers=DEFAULT_HEADERS).content)
            report = pdfplumber.open(filepath)
            obs['coa_pdf'] = f'{filename}.pdf'
        else:
            report = pdfplumber.open(doc)
            obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]

    raise NotImplementedError


# === Tests ===

from cannlytics.data.coas import CoADoc


# Initialize a COA parser.
parser = CoADoc()

# Parse a COA with cannabinoid-only testing.
url = 'https://kcalabs.qbench.net/pr/ee1b30d13037a54'
doc = '../../../tests/assets/coas/kca-labs/Tangie.pdf'

# Parse a COA with full QC testing.
# url = 'https://kcalabs.qbench.net/pr/c1f17c307ca6dc2'
# doc = '../../../tests/assets/coas/kca-labs/Elyxer-Gummies.pdf'

# Read the PDF.
report = pdfplumber.open(doc)
front_page = report.pages[0]

# TODO: Get the image.
# - Save the image locally from the image data.
# - Upload the image to Firebase Storage.
# - Create a dynamic link for the image (`sample_image_url`).
img = front_page.images[3]

# Get the front page text.
text = front_page.extract_text()
split = re.split('\d+ of \d+', text)
header = split[0]
body = split[-1].split('ND = Not Detected')[0]
lines = body.split('\n')
lines = [x for x in lines if x.strip()]

# Optional: Get the lab details from the header.
# block = header.split('\n')
# street = block[1].replace('https://kcalabs.com', '').strip()
# address_parts = block[2].split('Lic.#')
# obs['lab_license_number'] = address_parts[-1].strip()
# address = address_parts[0]

# TODO: Get the client details (if present).

# Future work: Get the moisture content and foreign matter.

# Iterate over lines to get sample details and cannabinoid results.
obs = {}
analyses = []
dates_tested = []
product_name = lines[0]
results = []
for line in lines[1:]:

    # Remove excess spaces from the lines.
    line_text = line.strip()

    # Skip nuisance rows.
    if line_text in coa_skip_values:
        continue

    # Get the sample details.
    # FIXME: Handle COAs with client details.
    elif ':' in line_text:
        fields = list(coa_sample_detail_fields.keys())
        for field in fields:
            if f'{field}:' in line_text:
                key = coa_sample_detail_fields[field]
                if obs.get(key):
                    continue
                detail = line_text.replace(f'{field}:', '').strip()
                secondary_detail = None
                for secondary_field in fields:
                    if secondary_field != field and f'{secondary_field}:' in detail:
                        parts = detail.split(f'{secondary_field}:')
                        secondary_detail = parts[-1].strip()
                        secondary_key = coa_sample_detail_fields[secondary_field]
                        obs[secondary_key] = secondary_detail
                        obs[key] = parts[0].strip()
                        continue
                if secondary_detail is None:
                    obs[key] = detail

    # Get the analyses and date tested.
    elif line_text.endswith('Tested'):
        parts = line_text.replace('Tested', '').strip().split(' ')
        date = pd.to_datetime(parts[-1])
        dates_tested.append(date)

    # Get the analysis method.
    elif ' by ' in line_text:
        parts = line_text.split(' by ')
        analysis = parts[0].lower()
        analysis = parser.analyses.get(analysis, analysis)
        method = parts[-1]
        obs[f'{analysis}_method'] = method

    # Get the total cannabinoids.
    elif line_text.startswith('Total') and not line_text.endswith('Internal Standard'):
        parts = line_text.split(' ')
        parts = [x for x in parts if x.strip()]
        key = snake_case(parts[0])
        if key == 'total':
            key = 'total_cannabinoids'
        key = parser.fields.get(key, key)
        obs[key] = convert_to_numeric(parts[1])

    # Get the cannabnoid results.
    else:
        parts = line_text.split(' ')
        parts = [x for x in parts if x.strip()]
        name = parts[0]
        analyte = snake_case(name)
        analyte = parser.analytes.get(analyte, analyte)
        columns = KCA_LABS_ANALYSES['cannabinoids']['columns']
        result = {
            'analysis': 'cannabinoids',
            'units': 'percent',
            'key': analyte,
        }
        for i, column in enumerate(columns):
            result[column] = convert_to_numeric(parts[i])
        if name == 'Total' or isinstance(result['name'], float):
            continue
        results.append(result)

# Get the results from the subsequent pages.
for page in report.pages[1:]:

    # Split at page number to remove the header,
    # then split at the note to remove the footer,
    # then remove the product name and sample details.
    text = page.extract_text()
    split = re.split('\d+ of \d+', text)
    body = split[-1].split('ND = Not Detected')[0]
    lines = body.split('\n')
    lines = [x for x in lines if x.strip()]
    lines = [x for x in lines if x != product_name and ':' not in x]

    # Get the analysis and analysis method.
    analysis_parts = lines[0].strip().split(' by ')
    analysis = snake_case(analysis_parts[0])
    analysis = parser.analyses.get(analysis, analysis)
    obs[f'{analysis}_method'] = analysis_parts[-1]
    analyses.append(analysis)

    # Get the columns.
    columns = lines[1].split(' ')
    columns = [x.lower() for x in columns if x and '(' not in x]
    columns = [x.replace('result', 'value').replace('analyte', 'name') for x in columns]
    units = lines[1].split('(')[-1].split(')')[0]
    if 'name' not in columns:
        if len(columns) == 6:
            columns = ['name'] + columns[:3] + ['name'] + columns[3:]
        else:
            columns.insert(0, 'name')

    # Compile the results.
    for line in lines[2:]:
        result = {'analysis': analysis, 'units': units}
        values = line.split(' ')

        # FIXME: If name is in the columns twice, then handle accordingly.
        # FIXME: Combine analyte names with spaces
        k = len(columns)
        if len(values) > k:
            half_point = round(k / 2)
            first_break = 1
            if k % 2 ==0 :
                first_break = 2
            second_break = - (half_point + 1)
            values = [' '.join(values[:first_break])] + values[first_break:half_point]
            values = [' '.join(values[half_point:second_break])] + values[second_break:]

        # Record the result values.
        for i, column in enumerate(columns):
            result[column] = convert_to_numeric(values[i])
            if column == 'name':
                analyte = snake_case(result['name'])
                analyte = parser.analytes.get(analyte, analyte)
                result['key'] = analyte
        results.append(result)

# Turn dates to ISO format.
date_columns = [x for x in obs.keys() if x.startswith('date')]
for date_column in date_columns:
    try:
        obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
    except:
        pass

# Ensure numeric columns are numeric.
for col in coa_numeric_columns:
    obs[col] = convert_to_numeric(obs[col])

# Finish data collection with a freshly minted sample ID.
date_tested = max(dates_tested).isoformat()
obs['analyses'] = analyses
obs['date_tested'] = date_tested
obs['product_name'] = product_name
obs['results'] = results

# TODO: Create `sample_hash` and `results_hash`.


obs['sample_id'] = create_sample_id(
    private_key=json.dumps(results),
    public_key=product_name,
    salt=date_tested,
)
obs['coa_parsed_at'] = datetime.now().isoformat()
obs['coa_algorithm_version'] = __version__
print({**KCA_LABS, **obs})


# Cannabis Data Science Exercise:
# Look at the Tangie relative to other Tangie test results from
# MCR Labs, SC Labs, and PSI Labs.



"""
CCRS Module | Cannlytics
Copyright (c) 2022-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 4/10/2022
Updated: 4/16/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Standard imports:
import os
from typing import Callable, List, Optional
from zipfile import ZipFile

# External imports:
import pandas as pd

# Internal imports:
from cannlytics.data import create_hash
from cannlytics.data.ccrs.constants import (
    CCRS_ANALYTES,
    CCRS_ANALYSES,
    CCRS_DATASETS,
)
from cannlytics.utils.utils import (
    camel_to_snake,
    convert_to_numeric,
    rmerge,
    sorted_nicely,
)


def anonymize(
        df: pd.DataFrame,
        columns: Optional[List[str]] = None,
    ) -> pd.DataFrame:
    """Anonymize a CCRS dataset."""
    if columns is None:
        columns = df.filter(regex=r'.*_by$|.*_By$', axis=1).columns
    for column in columns:
        df.loc[:, column] = df[column].astype(str).apply(create_hash)
    return df


def get_datafiles(
        data_dir: str,
        dataset: Optional[str] = '',
        desc: Optional[bool] = True,
    ) -> list:
    """Get all CCRS datafiles of a given type in a directory."""
    datafiles = sorted_nicely([
        os.path.join(data_dir, f, f, f + '.csv')
        for f in os.listdir(data_dir) if f.startswith(dataset)
    ])
    if desc:
        datafiles.reverse()
    return datafiles


def format_test_value(tests, compound, value_key='TestValue'):
    """Format a lab result test value from a DataFrame of tests."""
    value = tests.loc[(tests.key == compound)]
    try:
        return convert_to_numeric(value.iloc[0][value_key])
    except:
        return None


def find_detections(
        tests,
        analysis,
        analysis_key='analysis',
        analyte_key='key',
        value_key='value',
    ) -> List[str]:
    """Find compounds detected for a given analysis from given tests."""

    # Find detections in a list of results.
    if isinstance(tests, list):
        detected = []
        for test in tests:
            if test[analysis_key] == analysis:
                value = pd.to_numeric(test[value_key], errors='coerce')
                if value > 0:
                    detected.append(test[analyte_key])
        return detected

    # Find detections in a DataFrame.
    analysis_tests = tests.loc[tests[analysis_key] == analysis]
    if analysis_tests.empty:
        return []
    values = analysis_tests[value_key].apply(
        lambda x: pd.to_numeric(x, errors='coerce')
    )
    analysis_tests = analysis_tests.assign(**{value_key: values})
    detected = analysis_tests.loc[analysis_tests[value_key] > 0]
    if detected.empty:
        return []
    return detected[analyte_key].to_list()


def format_lab_results(
        df: pd.DataFrame,
        results: pd.DataFrame,
        item_key: Optional[str] = 'InventoryId',
        analysis_name: Optional[str] = 'TestName',
        analysis_key: Optional[str] = 'TestValue',
    ) -> pd.DataFrame:
    """Format CCRS lab results to merge into another dataset."""

    # Curate lab results.
    analyte_data = results[analysis_name].map(CCRS_ANALYTES).values.tolist()
    results = results.join(pd.DataFrame(analyte_data))
    results['type'] = results['type'].map(CCRS_ANALYSES)

    # Find lab results for each item.
    formatted = []
    item_ids = list(df[item_key].unique())
    for item_id in item_ids:
        item_results = results.loc[results[item_key].astype(str) == item_id]
        if item_results.empty:
            continue

        # Map certain test values.
        values = item_results.iloc[0].to_dict()
        [values.pop(key) for key in [analysis_name, analysis_key]]
        entry = {
            **values,
            'analyses': list(item_results['type'].unique()),
            'delta_9_thc': format_test_value(item_results, 'delta_9_thc'),
            'thca': format_test_value(item_results, 'thca'),
            'total_thc': format_test_value(item_results, 'total_thc'),
            'cbd': format_test_value(item_results, 'cbd'),
            'cbda': format_test_value(item_results, 'cbda'),
            'total_cbd': format_test_value(item_results, 'total_cbd'),
            'moisture_content': format_test_value(item_results, 'moisture_content'),
            'water_activity': format_test_value(item_results, 'water_activity'),
        }

        # Determine `status`.
        statuses = list(item_results['LabTestStatus'].unique())
        if 'Fail' in statuses:
            entry['status'] = 'Fail'
        else:
            entry['status'] = 'Pass'

        # Determine detected contaminants.
        entry['pesticides'] = find_detections(item_results, 'pesticides')
        entry['residual_solvents'] = find_detections(item_results, 'residual_solvent')
        entry['heavy_metals'] = find_detections(item_results, 'heavy_metals')

        # Record the lab results for the item.
        formatted.append(entry)

    # Return the lab results.
    return pd.DataFrame(formatted)


def merge_datasets(
        df: pd.DataFrame,
        datafiles: List[str],
        dataset: str,
        on: Optional[str] = 'id',
        target: Optional[str] = '',
        how: Optional[str] = 'left',
        sep: Optional[str] = '\t',
        validate: Optional[str] = 'm:1',
        dedupe: Optional[bool] = False,
        drop: Optional[dict] = None,
        rename: Optional[dict] = None,
        break_once_matched: Optional[bool] = True,
        on_bad_lines: Optional[str] = 'skip',
        string_columns: Optional[list] = [
            'IsDeleted', 'UnitWeightGrams',
            'TestValue', 'IsQuarantine'],
    ) -> pd.DataFrame:
    """Merge a supplemental CCRS dataset to an existing dataset.
    Note: Certain columns are treated as strings due to `ValueError`s.
    Lab results cannot be split between datafiles.
    """
    n = len(df)
    augmented = pd.DataFrame()
    fields = CCRS_DATASETS[dataset]['fields']
    parse_dates = CCRS_DATASETS[dataset]['date_fields']
    usecols = list(fields.keys()) + parse_dates
    dtype = {k: v for k, v in fields.items() if v != 'datetime64'}
    for column in string_columns:
        if dtype.get(column):
            dtype[column] = 'string'
    for datafile in datafiles:
        print(datafile)
        supplement = pd.read_csv(
            datafile,
            sep=sep,
            encoding='utf-16',
            engine='python',
            parse_dates=parse_dates,
            dtype=dtype,
            usecols=usecols,
            on_bad_lines=on_bad_lines,
        )
        if dedupe:
            supplement.drop_duplicates(subset=[on], inplace=True)
        if rename is not None:
            supplement.rename(rename, axis=1, inplace=True)
        if drop is not None:
            supplement.drop(drop, axis=1, inplace=True)
        if dataset == 'lab_results':
            supplement = format_lab_results(df, supplement)
        match = rmerge(
            df,
            supplement,
            on=on,
            how=how,
            validate=validate,
        )
        matched = match.loc[~match[target].isna()]
        augmented = pd.concat([augmented, matched], ignore_index=True)
        # FIXME: This is a hack to prevent the loop from running forever.
        if len(augmented) == n and break_once_matched:
            break
    return augmented


def save_dataset(
        data: pd.DataFrame,
        data_dir: str,
        name: str,
        ext: Optional[str] = 'xlsx',
        rows: Optional[float] = 1e6,
    ) -> None:
    """Save a curated dataset, determining the number of datafiles
    (1 million per file) and saving each shard of the dataset."""
    if not os.path.exists(data_dir): os.makedirs(data_dir)
    file_count = round((len(data) + rows) / rows)
    for i in range(0, file_count):
        start = int(0 + rows * i)
        stop = int(rows + rows * i)
        shard = data.iloc[start: stop, :]
        outfile = os.path.join(data_dir, f'{name}_{i}.{ext}')
        shard.to_excel(outfile, index=False)


def standardize_dataset(
        df: pd.DataFrame,
        rename_function: Optional[Callable] = camel_to_snake,
    ) -> pd.DataFrame:
    """Standardize a given DataFrame."""
    # Standardize column names.
    columns = {col: rename_function(col) for col in df.columns}
    df.rename(columns=columns, inplace=True)

    # Anonymize the data.
    df = anonymize(df)

    # Return the sorted data.
    return df.sort_index(axis=1)


def unzip_datafiles(
        data_dir: str,
        verbose: Optional[bool] = True,
    ) -> None:
    """Unzip all CCRS datafiles in a given directory."""
    zip_files = [f for f in os.listdir(data_dir) if f.endswith('.zip')]
    for zip_file in zip_files:
        filename = os.path.join(data_dir, zip_file)
        zip_dest = filename.rstrip('.zip')
        if not os.path.exists(zip_dest):
            os.makedirs(zip_dest)
        zip_ref = ZipFile(filename)
        zip_ref.extractall(zip_dest)
        zip_ref.close()
        os.remove(filename)
        if verbose:
            print('Unzipped:', zip_file)



"""
Utility Functions | Cannlytics
Copyright (c) 2021-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 2/5/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: This module contains general Cannlytics utility functions.
"""
# Standard imports.
from base64 import b64encode, decodebytes
from datetime import datetime, timedelta
import json
import os
from re import split, sub, findall
import secrets
from typing import Any, Callable, List, Optional, Tuple
from zipfile import ZipFile
try:
    from zoneinfo import ZoneInfo
except ImportError:
    print('Operating with Python < 3.9 is not recommended.')

# External imports.
from dateutil import parser, relativedelta
from fredapi import Fred
from pandas import ExcelWriter, merge, NaT, to_datetime
from pandas.tseries.offsets import MonthEnd
import requests

# Internal imports.
try:
    from cannlytics.utils.constants import (
        RANDOM_STRING_CHARS,
        state_time_zones,
    )
except ImportError:
    print('Failed to load constants.')


#-----------------------------------------------------------------------
# String utilities.
#-----------------------------------------------------------------------

def camelcase(string: str) -> str:
    """Turn a given string to CamelCase.
    Args:
        string (str): A given string to turn to CamelCase.
    Returns:
        (str): A string in CamelCase.
    """
    key = string.replace('&', 'and')
    key = key.replace('%', 'percent')
    key = key.replace('#', 'number')
    key = key.replace('$', 'dollars')
    key = key.replace('/', 'to')
    key = key.replace('.', '_')
    key = ''.join(x for x in key.title() if not x.isspace())
    key = key.replace('_', '').replace('-', '')
    return key


def camel_to_snake(string: str) -> str:
    """Turn a camel-case string to a snake-case string.
    This function handles CamelCase better than `snake_case`.
    The function does not do well with all caps, e.g. "APP_ID".
    Args:
        string (str): The string to convert to snake-case.
    Returns:
        (str): Returns the string in snake_case.
    """
    return sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


def kebab_case(string: str) -> str:
    """Turn a string into a kebab-case string."""
    key = string.replace(' ', '-')
    key = key.replace('&', 'and')
    key = key.replace('%', 'percent')
    key = key.replace('#', 'number')
    key = key.replace('$', 'dollars')
    key = key.replace('/', 'to')
    key = key.replace(r'\\', '-').lower()
    key = sub('[!@#$%^&*()[]{};:,./<>?\|`~-=+]', ' ', key)
    keys = findall(r'[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+', key)
    return '-'.join(map(str.lower, keys))


def format_billions(value: float, pos: Optional[int] = None) -> str: #pylint: disable=unused-argument
    """The two args are the value and tick position."""
    return '%1.1fB' % (value * 1e-9)


def format_millions(value: float, pos: Optional[int] = None) -> str: #pylint: disable=unused-argument
    """The two args are the value and tick position."""
    return '%1.1fM' % (value * 1e-6)


def format_thousands(value: float, pos: Optional[int] = None) -> str: #pylint: disable=unused-argument
    """The two args are the value and tick position."""
    return '%1.0fK' % (value * 1e-3)


def get_keywords(string: str) -> List[str]:
    """Get keywords for a given string.
    Args:
        string (str): A string to get keywords for.
    Returns:
        (list): A list of keywords.
    """
    keywords = string.lower().split(' ')
    keywords = [x.strip() for x in keywords if x]
    keywords = list(set(keywords))
    return keywords


def get_random_string(length, allowed_chars=RANDOM_STRING_CHARS):
    """
    Return a securely generated random string.

    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)

    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits

    Copyright (c) Django Software Foundation and individual contributors.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
           this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
           notice, this list of conditions and the following disclaimer in the
           documentation and/or other materials provided with the distribution.

        3. Neither the name of Django nor the names of its contributors may be used
           to endorse or promote products derived from this software without
           specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    return ''.join(secrets.choice(allowed_chars) for i in range(length))


def sentence_case(s):
    """Format a string as a sentence.
    Author: Zizouz212 https://stackoverflow.com/a/39969233/5021266
    License: CC BY-SA 3.0 https://creativecommons.org/licenses/by-sa/3.0/
    """
    return '. '.join(i.capitalize() for i in s.split('. ')).strip()


REPLACEMENTS = [
    {'text': ' ', 'key': '_'},
    {'text': '&', 'key': 'and'},
    {'text': '%', 'key': 'percent'},
    {'text': '#', 'key': 'number'},
    {'text': '$', 'key': 'dollars'},
    {'text': '/', 'key': 'to'},
    {'text': 'α', 'key': 'alpha'},
    {'text': 'β', 'key': 'beta'},
    {'text': 'γ', 'key': 'gamma'},
    {'text': 'Δ', 'key': 'delta'},
    {'text': 'δ', 'key': 'delta'},
]

def snake_case(string: str) -> str:
    """Turn a given string to snake case.
    Handles CamelCase, replaces known special characters with
    preferred namespaces, replaces spaces with underscores,
    and removes all other nuisance characters.
    Args:
        string (str): The string to turn to snake case.
    Returns:
        (str): A snake case string.
    """
    key = string.replace(r'\\', '_').lower()
    for x in REPLACEMENTS:
        key = key.replace(x['text'], x['key'])
    key = sub('[!@#$%^&*()[]{};:,./<>?\|`~-=+]', ' ', key)
    keys = findall(r'[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+', key)
    return '_'.join(map(str.lower, keys))


def strip_whitespace(string: str) -> str:
    """Strip whitespace from a string."""
    return string.replace('\n', '').strip()


#-----------------------------------------------------------------------
# Number utilities.
#-----------------------------------------------------------------------

def convert_to_numeric(string: str, strip: Optional[str] = False) -> Any:
    """Convert a string to numeric, optionally replacing non-numeric
    characters.
    Args:
        string (str): The string to attempt to parse to a number.
    Returns:
        (float): Returns either the original or the parsed number.
    """
    if strip:
        s = sub('[^\d\.]', '', string)
    else:
        s = string
    try:
       return float(s)
    except (TypeError, ValueError):
        return s


#-----------------------------------------------------------------------
# List utilities.
#-----------------------------------------------------------------------

def sandwich_list(a) -> list:
    """Create a range that cycles from start to the end to the middle.
    Credit: Norman <https://stackoverflow.com/a/36533868/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    Args:
        a (str): The iterable to sandwich.
    Returns:
        (list): The sandwich index.
    """
    return [a[-i//2] if i % 2 else a[i//2] for i in range(len(a))]


def sorted_nicely(unsorted_list: List[str]) -> List[str]:
    """Sort the given iterable in a way that humans would expect.
    Credit: Mark Byers <https://stackoverflow.com/a/2669120/5021266>
    License: CC BY-SA 2.5 <https://creativecommons.org/licenses/by-sa/2.5/>
    """
    convert = lambda text: int(text) if text.isdigit() else text
    alpha = lambda key: [convert(c) for c in split('([0-9]+)', key)]
    return sorted(unsorted_list, key=alpha)


def split_list(a_list: list, at_index: Optional[int] = None) -> Tuple:
    """Split a list in half or at a given index.
    Credit: Jason Coon <https://stackoverflow.com/a/752330/5021266>
    License: CC BY-SA 4.0 <https://creativecommons.org/licenses/by-sa/4.0/>
    """
    if at_index:
        half = at_index
    else:
        half = len(a_list)//2
    return a_list[:half], a_list[half:]


#-----------------------------------------------------------------------
# Dictionary utilities.
#-----------------------------------------------------------------------

def clean_dictionary(data: dict, function: Callable = snake_case) -> dict:
    """Format dictionary keys with given function, snake case by default.
    Args:
        d (dict): A dictionary to clean.
        function (function): A function to apply to each key.
    Returns:
        (dict): Returns the input dictionary with a function applied to the keys.
    """
    return {function(k): v for k, v in data.items()}


def clean_nested_dictionary(data: dict, function: Callable = snake_case) -> dict:
    """Format nested (at most 2 levels) dictionary keys with a given function,
    snake case by default. Handles list of dictionaries.
    Args:
        d (dict): A dictionary to clean, allowing dictionaries as values.
        function (function): A function to apply to each key.
    Returns:
        (dict): Returns the input dictionary with cleaned keys.
    """
    clean = clean_dictionary(data, function)
    for k, value in clean.items():
        if isinstance(value, list):
            x = []
            for v in value:
                try:
                    x.append(clean_nested_dictionary(v, function))
                except:
                    x.append(v)
            clean[k] = x
        elif isinstance(value, dict):
            try:
                clean[k] = clean_nested_dictionary(value, function)
            except AttributeError:
                pass
    return clean


def remove_dict_fields(data: dict, fields: List[str]) -> dict:
    """Remove multiple keys from a dictionary.
    Args:
        data (dict): The dictionary to clean.
        fields (list): A list of keys (str) to remove.
    Returns:
        (dict): Returns the dictionary with the keys removed.
    """
    for key in fields:
        if key in data:
            del data[key]
    return data


def remove_dict_nulls(data: dict) -> dict:
    """Return a shallow copy of a dictionary with all `None` values excluded.
    Args:
        data (dict): The dictionary to reduce.
    Returns:
        (dict): Returns the dictionary with the keys with
            null values removed.
    """
    return {k: v for k, v in data.items() if v is not None}


def update_dict(context: dict, function: Callable = camel_to_snake, **kwargs) -> dict:
    """Update dictionary with keyword arguments.
    Args:
        context (dict): A dictionary of context to update.
        function (function): Function to apply to final dictionary keys.
        kwargs (*args): Key/value arguments to update in the dictionary.
    Returns:
        (dict): Returns the dictionary with updated keys.
    """
    entry = {}
    for key in kwargs:
        entry[key] = kwargs[key]
    data = {
        **clean_nested_dictionary(context, function),
        **clean_nested_dictionary(entry, function)
    }
    return data


#-----------------------------------------------------------------------
# DataFrame utilities.
#-----------------------------------------------------------------------

def clean_column_strings(data: Any, column: str) -> Any:
    """Clean the column names of a given DataFrame.
    Args:
        data (DataFrame): A DataFrame with any column names.
        column (str): The column of the DataFrame to clean.
    Returns:
        (DataFrame): A DataFrame with snake_case column names.
    """
    data[column] = data[column].str.strip()
    data[column] = data[column].str.rstrip('.)]')
    data[column] = data[column].str.replace('%', 'percent', regex=True)
    data[column] = data[column].str.replace('#', 'number', regex=True)
    data[column] = data[column].str.replace('[/,]', '_', regex=True)
    data[column] = data[column].str.replace('[.,(,)]', '', regex=True)
    data[column] = data[column].str.replace("\'", '', regex=True)
    data[column] = data[column].str.replace("[[]", '_', regex=True)
    data[column] = data[column].str.replace(r"[]]", '', regex=True)
    data[column] = data[column].str.replace('β', 'beta', regex=True)
    data[column] = data[column].str.replace('Δ', 'delta', regex=True)
    data[column] = data[column].str.replace('δ', 'delta', regex=True)
    data[column] = data[column].str.replace('α', 'alpha', regex=True)
    data[column] = data[column].str.replace('__', '', regex=True)
    return data


def dump_column(x):
    """Turn a column from JSON to dictionaries, handling errors."""
    try:
        return json.loads(x)
    except:
        return None


def end_of_period_timeseries(data: Any, period: Optional[str] = 'M') -> Any:
    """Convert a DataFrame from beginning-of-the-period to
    end-of-the-period timeseries.
    Args:
        data (DataFrame): The DataFrame to adjust timestamps.
        period (str): The period of the time series, monthly "M" by default.
    Returns:
        (DataFrame): The adjusted DataFrame, with end-of-the-month timestamps.
    """
    data.index = data.index.to_period(period).to_timestamp(period)
    return data


def nonzero_columns(data):
    """Return the non-zero column names of a DataFrame."""
    nonzero = (data != 0).any()
    return data.columns[nonzero].to_list()


def nonzero_rows(data):
    """Return the non-zero row keys of a DataFrame."""
    nonzero = (data != 0)
    return nonzero.index[nonzero].to_list()


def combine_columns(
        df: Any,
        new_key: str,
        old_key: str,
        drop: Optional[bool] = True,
    ):
    """Combine two numeric columns of a DataFrame.
    Args:
        df (DataFrame): The DataFrame with the columns.
        new_key (str): The column to have values filled.
        old_key (str): The column to use to fill values.
        drop (bool): Whether or not to drop the old column, the default
            is `True` (optional).
    Returns:
        (DataFrame): Returns the DataFrame with combined columns.
    """
    df.loc[df[new_key] == 0, new_key] = df[old_key]
    df[new_key].fillna(df[old_key], inplace=True)
    if drop:
        df.drop(columns=[old_key], inplace=True)
    return df


def reorder_columns(df: Any, columns: list) -> Any:
    """Re-order a DataFrame given a specific order of columns.
    Remaining columns will be appended to the end of the DataFrame.
    Args:
        data (DataFrame): A DataFrame to be re-ordered.
        columns (list): A list of columns in desired order.
    Returns:
        (DataFrame): Returns the re-ordered DataFrame
    """
    rdf = df.copy()
    excluded = list([a for a in rdf.columns if a not in columns])
    ordered_columns = (columns + excluded)
    rdf = rdf.reindex(columns=sorted(rdf.columns))
    rdf = rdf.reindex(columns=ordered_columns)
    return rdf


def reverse_dataframe(data: Any) -> Any:
    """Reverse the ordering of a DataFrame.
    Args:
        data (DataFrame): A DataFrame to re-order.
    Returns:
        (DataFrame): The re-ordered DataFrame.
    """
    return data[::-1].reset_index(drop=True)


def sum_columns(
        df: Any,
        new_key: str,
        columns: list,
        drop: Optional[bool] = True,
    ):
    """Sum multiple numeric columns of a DataFrame.
    Args:
        df (DataFrame): The DataFrame with the columns.
        new_key (str): The column to have values filled.
        columns (list): The columns to use to fill values.
        drop (bool): Whether or not to drop the old column, the default
            is `True` (optional).
    Returns:
        (DataFrame): Returns the DataFrame with summed columns.
    """
    df[new_key] = 0
    for column in columns:
        df[new_key] += df[column].fillna(0)
    if drop:
        df.drop(columns=columns, inplace=True)
    return df


def rmerge(left, right, **kwargs):
    """Perform a merge using pandas with optional removal of overlapping
    column names not associated with the join.

    Though I suspect this does not adhere to the spirit of pandas merge
    command, I find it useful because re-executing IPython notebook cells
    containing a merge command does not result in the replacement of existing
    columns if the name of the resulting DataFrame is the same as one of the
    two merged DataFrames, i.e. data = pa.merge(data, df). I prefer
    this command over pandas df.combine_first() method because it has more
    flexible join options.

    The column removal is controlled by the 'replace' flag which is
    'left' (default) or 'right' to remove overlapping columns in either the
    left or right DataFrame. If 'replace' is set to None, the default
    pandas behavior will be used. All other parameters are the same
    as pandas merge command.

    Author: Michelle Gill
    Source: https://gist.github.com/mlgill/11334821

    Examples
    --------
    >>> left       >>> right
       a  b   c       a  c   d
    0  1  4   9    0  1  7  13
    1  2  5  10    1  2  8  14
    2  3  6  11    2  3  9  15
    3  4  7  12

    >>> rmerge(left, right, on='a')
       a  b  c   d
    0  1  4  7  13
    1  2  5  8  14
    2  3  6  9  15

    >>> rmerge(left, right, on='a', how='left')
       a  b   c   d
    0  1  4   7  13
    1  2  5   8  14
    2  3  6   9  15
    3  4  7 NaN NaN

    >>> rmerge(left, right, on='a', how='left', replace='right')
       a  b   c   d
    0  1  4   9  13
    1  2  5  10  14
    2  3  6  11  15
    3  4  7  12 NaN

    >>> rmerge(left, right, on='a', how='left', replace=None)
       a  b  c_x  c_y   d
    0  1  4    9    7  13
    1  2  5   10    8  14
    2  3  6   11    9  15
    3  4  7   12  NaN NaN
    """

    # Function to flatten lists from http://rosettacode.org/wiki/Flatten_a_list#Python
    def flatten(lst):
        return sum(([x] if not isinstance(x, list) else flatten(x) for x in lst), [])

    # Set default for removing overlapping columns in "left" to be true
    myargs = {'replace':'left'}
    myargs.update(kwargs)

    # Remove the replace key from the argument dict to be sent to
    # pandas merge command
    kwargs = {k:v for k, v in myargs.items() if k != 'replace'}

    if myargs['replace'] is not None:
        # Generate a list of overlapping column names not associated with the join
        skip_cols = set(flatten([v for k, v in myargs.items() if k in ['on', 'left_on', 'right_on']]))
        left_cols = set(left.columns)
        right_cols = set(right.columns)
        drop_cols = list((left_cols & right_cols).difference(skip_cols))

        # Remove the overlapping column names from the appropriate DataFrame
        if myargs['replace'].lower() == 'left':
            left = left.copy().drop(drop_cols, axis=1)
        elif myargs['replace'].lower() == 'right':
            right = right.copy().drop(drop_cols, axis=1)

    return merge(left, right, **kwargs)


def set_training_period(series: Any, date_start: str, date_end: str) -> Any:
    """Helper function to restrict a series to the desired
    training time period.
    Args:
        series (Series): The series to clean.
        date_start (str): An ISO date to mark the beginning of the training period.
        date_end (str): An ISO date to mark the end of the training period.
    Returns
        (Series): The series restricted to the desired time period.
    """
    return series.loc[
        (series.index >= to_datetime(date_start)) & \
        (series.index < to_datetime(date_end))
    ]


def to_excel_with_style(
        df: Any,
        file_name: str,
        index: Optional[bool] = False,
        sheet_name: Optional[str] = 'Sheet1',
        style: Optional[dict] = None,
    ):
    """Save a DataFrame to Excel with no style.
    Args:
        df (DataFrame): The data to save.
        file_name (str): The name of the file.
        index (bool): Whether to include the index, `False` by default (optional).
        sheet_name (str): The name for the worksheet, `Sheet1` by default (optional).
        style (dict): The style to apply to the headers (optional). Applies
            a light green background with an underline by default.
    """
    if style is None:
        style = {'bottom': 1, 'bg_color': '#EBF1DE'}
    writer = ExcelWriter(file_name, engine='xlsxwriter')
    df.to_excel(writer, index=index, sheet_name=sheet_name, startrow=1, header=False)
    worksheet = writer.sheets[sheet_name]
    workbook = writer.book
    bold = workbook.add_format(style)
    for idx, val in enumerate(df.columns):
        worksheet.write(0, idx, val, bold)
    writer.close()


#-----------------------------------------------------------------------
# Time utilities.
#-----------------------------------------------------------------------

def convert_month_year_to_date(x):
    """Convert a month, year series to datetime. E.g. `'April 2022'`."""
    try:
        return datetime.strptime(x.replace('.0', ''), '%B %Y')
    except:
        return NaT


def end_of_month(value: datetime) -> str:
    """Format a datetime as an ISO formatted date at the end of the month.
    Args:
        value (datetime): A datetime value to transform into an ISO date.
    Returns:
        (str): An ISO formatted date.
    """
    month = value.month
    if month < 10:
        month = f'0{month}'
    year = value.year
    day = value + MonthEnd(0)
    return f'{year}-{month}-{day.day}'


def end_of_year(value: datetime) -> str:
    """Format a datetime as an ISO formatted date at the end of the year.
    Args:
        value (datetime): A datetime value to transform into an ISO date.
    Returns:
        (str): An ISO formatted date.
    """
    return f'{value.year}-12-31'


def format_iso_date(date: str, sep: Optional[str] = '/') -> str:
    """Format a human-written date into an ISO formatted date.
    Ags:
        date (str): A human-written date.
        sep (str): The separating character, '/' by default (optional).
    Returns:
        (str): An ISO formatted date string.
    """
    mm, dd, yyyy = tuple(date.split(sep))
    if len(mm) == 1:
        mm = f'0{mm}'
    if len(dd) == 1:
        dd = f'0{dd}'
    if len(yyyy) == 2:
        yyyy = f'20{yyyy}'
    return '-'.join([yyyy, mm, dd])


def get_date_range(start_date: str, end_date: str):
    """Generates a list of tuples containing ISO-formatted date strings
    for each consecutive day in a given range of dates.
    Args:
        start_date (str): The start date in the format '%Y-%m-%d'.
        end_date (str): The end date in the format '%Y-%m-%d'.
    Returns:
        The output will be a list of tuples containing date strings in
        format '%Y-%m-%d'
    Credit: ChatGPT <https://chat.openai.com/chat>
    """
    start_date = datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.strptime(end_date, '%Y-%m-%d')
    delta = timedelta(days=1)
    date_range = []
    while start_date <= end_date:
        start = start_date.strftime('%Y-%m-%d')
        end = (start_date + delta).strftime('%Y-%m-%d')
        date_range.append((start, end))
        start_date += delta
    return date_range


def get_timestamp(
        date: Optional[str] = None,
        past: Optional[int] = 0,
        future: Optional[int] = 0,
        zone: Optional[str] = 'utc'
) -> str:
    """Get an ISO formatted timestamp.
    Args:
        date (str): An optional date to pin the timestamp to a specific time.
        past (int): Number of minutes in the past to get a timestamp.
        future (int): Number of minutes into the future to get a timestamp.
        zone (str): A specific timezone or US state abbreviation, e.g. CA.
    Returns:
        (str): An ISO formatted date/time string in the specified time zone,
            UTC by default.
    """
    time_zone = state_time_zones.get(str(zone).upper(), zone)
    if date:
        now = parser.parse(date).replace(tzinfo=ZoneInfo(time_zone))
    elif time_zone is None:
        now = datetime.now()
    else:
        now = datetime.now(ZoneInfo(time_zone))
    now += timedelta(minutes=future)
    now -= timedelta(minutes=past)
    if time_zone is None:
        return now.isoformat()[:19]
    else:
        return now.isoformat()


def months_elapsed(start, end):
    """Calculate the months elapsed between two datetimes,
    returning 0 if a negative time span.
    """
    diff = relativedelta.relativedelta(end, start)
    time_span = diff.months + diff.years * 12
    return time_span if time_span > 0 else 0


#-----------------------------------------------------------------------
# File utilities.
#-----------------------------------------------------------------------

def decode_pdf(data: str, destination: str):
    """Save an base-64 encoded string as a PDF.
    Args:
        data (str): Base-64 encoded string representing a PDF.
        destination (str): The destination for the PDF file.
    """
    bits = decodebytes(data)
    with open(destination, 'wb') as pdf:
        pdf.write(bits)


def encode_pdf(filename: str) -> str:
    """Open a PDF file in binary mode.
    Args:
        filename (str): The full file path of a PDF to encode.
    Returns:
        (str): A string encoded in base-64.
    """
    with open(filename, 'rb') as pdf:
        return b64encode(pdf.read())


def get_directory_files(
        target_dir: str,
        file_type: Optional[str] = 'pdf',
    ) -> list:
    """Get all of the files of a specified type in a given directory.
    Args:
        target_dir (str): The directory containing the files.
        file_type (str): The file type extension, 'pdf' by default (optional).
    Returns:
    """
    return [
        os.path.join(target_dir, f) for f in os.listdir(target_dir) \
        if os.path.isfile(os.path.join(target_dir, f)) \
        and f.endswith(file_type)
    ]


def get_blocks(files, size=65536):
    """Get a block of a file by the given size."""
    while True:
        block = files.read(size)
        if not block: break
        yield block


def get_number_of_lines(
        file_name: str,
        encoding: Optional[str] = 'utf-16',
        errors: Optional[str] = 'ignore',
        verbose: Optional[bool] = False,
    ) -> int:
    """
    Read the number of lines in a large file, such as a .csv or .tsv file.
    Args:
        file_name (str): The file to count the lines, expecting a .csv file.
        encoding (str): The encoding of the file, "utf-16" by default (optional).
        errors (str): How to handle the errors, "ignore" by default (optional).
        verbose (bool): Whether or not to print the count, `False` by
            default (optional).
    Returns:
        (int): The number of lines of the file.
    Credit: glglgl, SU3 <https://stackoverflow.com/a/9631635/5021266>
    License: CC BY-SA 3.0 <https://creativecommons.org/licenses/by-sa/3.0/>
    """
    with open(file_name, 'r', encoding=encoding, errors=errors) as f:
        count = sum(bl.count('\n') for bl in get_blocks(f))
        if verbose:
            print('Number of rows:', count)
        return count


def download_file_from_url(url, destination='', ext='', file_name = None):
    """Download a file from a URL to a given directory.
    Author: H S Umer farooq <https://stackoverflow.com/a/53153505>
    License: CC BY-SA 4.0 https://creativecommons.org/licenses/by-sa/4.0/
    """
    get_response = requests.get(url, stream=True)
    if file_name is None:
        file_name = snake_case(url.split('/')[-1])
    if not file_name.endswith(ext):
        file_name = file_name + ext
    file_path = os.path.join(destination, file_name)
    with open(file_path, 'wb') as f:
        for chunk in get_response.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    return file_path


def unzip_files(zip_dir, extension='.zip'):
    """Unzip all files in a specified folder. Alternatively,
    pass a .zip file to extract that file.
    Author: nlavr https://stackoverflow.com/a/69101930
    License: CC BY-SA 4.0 https://creativecommons.org/licenses/by-sa/4.0/
    """
    single = ''
    zip_destinations = []
    if isinstance(zip_dir, str):
        parts = zip_dir.split('/')
        single = parts[-1]
        doc_dir = '/'.join(parts[:-1])
        if doc_dir == '':
            doc_dir = '.'
    else:
        doc_dir = zip_dir
    for item in os.listdir(doc_dir):
        if single and single not in item:
            continue
        abs_path = os.path.join(doc_dir, item)
        if item.endswith(extension):
            file_name = os.path.abspath(abs_path)
            zip_dest = os.path.join(doc_dir, item)
            if not os.path.exists(zip_dest):
                os.makedirs(zip_dest)
            zip_ref = ZipFile(file_name)
            zip_ref.extractall(zip_dest)
            zip_ref.close()
            os.remove(file_name)
            zip_destinations.append(zip_dest)
        elif os.path.isdir(abs_path):
            unzip_files(abs_path)
    return zip_destinations



"""
Cannlytics LIMS Initialization | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 11/6/2021
"""

# TODO: Decide which packages to make readily available.
# from .calculations import calculate_results
# from .certificates import (
#     create_coa,
#     delete_coa,
#     email_coas,
#     text_coas,
#     send_coas,
#     get_coa_urls,
#     merge_coas,
#     review_coa,
#     approve_coa,
#     upload_coa,
#     upload_coa_template,
# )
# from .data import (
#     import_data_model,
#     import_measurements,
# )
# from .results import (
#     post_results,
#     release_results,
#     send_results,
# )
# from .transfers import (
#     receive_transfer,
# )



"""
Cannlytics Models Initialization | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 11/6/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""

from .analysis import Analysis
from .analyte import Analyte
from .api_key import APIKey
from .area import Area
from .batch import Batch
from .calculation import Calculation
from .cannabis_license import CannabisLicense
from .certificate import Certificate
from .contact import Contact
from .dataset import DataSet
from .instrument import Instrument
from .inventory_type import InventoryType
from .invoice import Invoice
from .item import Item
from .measurement import Measurement
from .organization import Organization
from .price import Price
from .project import Project
from .regulation import Regulation
from .report import Report
from .sample import Sample
from .settings import UserSettings, OrganizationSettings
from .transfer import Transfer
from .user import User
from .workflow import Workflow

__all__ = [
    'Analysis',
    'Analyte',
    'APIKey',
    'Area',
    'Batch',
    'Calculation',
    'CannabisLicense',
    'Certificate',
    'Contact',
    'DataSet',
    'Instrument',
    'InventoryType',
    'Invoice',
    'Item',
    'Measurement',
    'Organization',
    'Price',
    'Project',
    'Regulation',
    'Report',
    'Sample',
    'UserSettings',
    'OrganizationSettings',
    'Transfer',
    'User',
    'Workflow',
]



"""
Parse Green Leaf Lab CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/23/2022
Updated: 9/7/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse a Green Leaf Labs CoA PDF.

Data Points:

    ✓ analyses
    - {analysis}_method
    ✓ {analysis}_status
    ✓ batch_size
    x coa_urls
    ✓ date_harvested
    ✓ date_tested
    ✓ date_received
    ✓ date_sampled
    ✓ distributor
    ✓ distributor_address
    ✓ distributor_street
    ✓ distributor_city
    ✓ distributor_state
    ✓ distributor_zipcode
    ✓ distributor_license_number
    x images
    ✓ image_data
    x lab_results_url
    ✓ metrc_lab_id
    ✓ metrc_source_id
    ✓ producer
    ✓ producer_address
    ✓ producer_street
    ✓ producer_city
    ✓ producer_state
    ✓ producer_zipcode
    ✓ producer_license_number
    ✓ product_name
    ✓ product_type
    ✓ results
    ✓ sample_weight
    ✓ sample_size
    ✓ sampling_method
    ✓ status
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_terpenes
    ✓ sample_id (generated)
    - strain_name (augmented)
    ✓ lab_id
    ✓ lab
    ✓ lab_image_url
    ✓ lab_license_number (should be dynamic)
    ✓ lab_address (should be dynamic)
    ✓ lab_street (should be dynamic)
    ✓ lab_city (should be dynamic)
    ✓ lab_county (augmented)
    ✓ lab_state (should be dynamic)
    ✓ lab_zipcode (should be dynamic)
    ✓ lab_phone (should be dynamic)
    - lab_email
    ✓ lab_website
    ✓ lab_latitude (augmented) (should be dynamic)
    ✓ lab_longitude (augmented) (should be dynamic)

"""
# Standard imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from typing import Any

# External imports.
import pandas as pd
import pdfplumber

# Internal imports.
from cannlytics.data.data import create_sample_id
from cannlytics.utils.constants import STANDARD_UNITS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)


# It is assumed that the lab has the following details.
# Future work: Make this dynamic to handle multiple lab locations.
# E.g. Green Leaf Lab has a California and an Oregon location.
GREEN_LEAF_LAB = {
    'coa_algorithm': 'greenleaflab.py',
    'coa_algorithm_entry_point': 'parse_green_leaf_lab_pdf',
    'lims': 'Green Leaf Lab',
    'lab': 'Green Leaf Lab',
    'lab_image_url': 'https://cdn-djjmk.nitrocdn.com/MuWSCTBsUZpIUufaWqGQkErSrYFMxIqD/assets/static/optimized/rev-a199899/wp-content/uploads/2018/12/greenleaf-logo.png',
    'lab_license_number': ' C8-0000078-LIC', # <- Make dynamic.
    'lab_address': '251 Lathrop Way Suites D&E Sacramento, CA 95815', # <- Make dynamic.
    'lab_street': '251 Lathrop Way Suites D&E', # <- Make dynamic.
    'lab_city': 'Sacramento', # <- Make dynamic.
    'lab_county': 'Sacramento', # <- Make dynamic.
    'lab_state': 'CA', # <- Make dynamic.
    'lab_zipcode': '95815', # <- Make dynamic.
    'lab_phone': '916-924-5227', # <- Make dynamic.
    'lab_email': '', # <- Make dynamic.
    'lab_website': 'https://greenleaflab.org/',
    'lab_latitude': 38.596060, # <- Make dynamic.
    'lab_longitude': -121.459870, # <- Make dynamic.
}

# It is assumed that there are the following analyses on each CoA.
GREEN_LEAF_LAB_ANALYSES = {
    'cannabinoids': {
        'name': 'Potency Analysis by HPLC',
        'columns': ['name', 'lod', 'loq', 'value', 'mg_g'],
    },
    'pesticides': {
        'name': 'Pesticide Analysis by GCMS/LCMS',
        'columns': ['name', 'value', 'limit', 'lod', 'loq', 'units'],
        'double_column': True,
    },
    'water_activity': {
        'name': 'Water Activity by Aqua Lab',
        'columns': ['name', 'value', 'units', 'lod', 'loq'],
    },
    'moisture_content': {
        'name': 'Moisture by Moisture Balance',
        'columns': ['name', 'value', 'units'],
    },
    'terpenes': {
        'name': 'Terpene Analysis by GCMS',
        'columns': ['name', 'value', 'mg_g'],
        'double_column': True,
    },
    'heavy_metals': {
        'name': 'Metals Analysis by ICPMS',
        'columns': ['name', 'value', 'limit', 'lod', 'loq', 'units'],
    },
    'mycotoxins': {
        'name': 'Mycotoxins by LCMSMS',
        'columns': ['name', 'value', 'limit', 'loq', 'loq', 'units'],
    },
    'microbials': {
        'name': 'Microbials by PCR',
        'columns': ['name', 'status', 'limit', 'lod', 'loq', 'units', 'value'],
    },
    'foreign_matter': {
        'name': 'Filth and Foreign Material Inspection by Magnification',
        'columns': ['name', 'status', 'limit', 'lod', 'loq', 'units'],
    },
}

# It is assumed that the CoA has the following parameters.
GREEN_LEAF_LAB_COA = {
    'coa_qr_code_index': None,
    'coa_page_area': '(0, 198, 612, 693)',
    'coa_distributor_area': '(0, 79.2, 244.8, 142.56)',
    'coa_producer_area': '(244.8, 79.2, 612, 142.56)',
    'coa_sample_details_area': '(0, 126.72, 612, 205.92)',
    'coa_sample_detail_fields': [
        'Test RFID',
        'Source RFID',
        'Lab Sample ID',
        'Sampling Method/SOP',
        'Source Batch ID',
        'Matrix',
        'Batch Size',
        'Sample Size',
        'Date Sampled',
        'Date Received',
        'Harvest/Processing Date',
        'Product Density',
    ],
    'coa_skip_values': [
        'Date/Time',
        'Analysis Method',
        'Analyte',
        'ND - Compound not detected',
        '<LOQ - Results below the Limit of Quantitation',
        'Results above the Action Level',
        'Sesquiterpenes',
        'Monoterpenes',
    ],
    # TODO: Make this field obsolete.
    'coa_replacements': [
        {'text': '< LOQ', 'key': '<LOQ'},
        {'text': '< LOD', 'key': '<LOD'},
        {'text': 'No detection in 1 gram', 'key': 'ND'},
        {'text': 'Ocimene isomer II', 'key': 'beta-ocimene'},
        {'text': 'Ocimene isomer I', 'key': 'alpha-ocimene'},
        {'text': 'p-Mentha-1,5-diene', 'key': 'p-Mentha-1-5-diene'},
        {'text': 'Methyl parathion', 'key': 'Methyl-parathion'},
        {'text': 'Caryophyllene Oxide', 'key': 'beta_caryophyllene_oxide'},
    ],
}


def augment_analyte_result(
        analytes: dict,
        result: dict,
        columns: list,
        parts: list,
    ) -> dict:
    """Quickly augment / add keys and values to an analyte result.
    Args:
        analytes (dict): A map of analytes to standard analytes.
        result (dict): A dictionary of results.
        columns (list): A list of keys to add to the result.
        parts (list): A list of values to add to the result.
    Returns:
        (dict): The updated result.
    """
    r = result.copy()
    if len(parts) > len(columns):
        break_point = len(parts) - len(columns) + 1
        name = ' '.join(parts[:break_point])
        analyte = snake_case(name)
        analyte = analytes.get(analyte, analyte)
        r['name'] = name
        r['key'] = analyte
        for i, part in enumerate(parts[break_point:]):
            r[columns[i + 1]] = convert_to_numeric(part)
    else:
        for i, part in enumerate(parts):
            if i == 0:
                analyte = snake_case(part)
                analyte = analytes.get(analyte, analyte)
                r['name'] = part
                r['key'] = analyte
            else:
                r[columns[i]] = convert_to_numeric(part)
    return r


def parse_green_leaf_lab_pdf(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a Green Leaf Lab CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    # Get the lab / LIMS analyses and CoA parameters.
    obs = {}
    lab_analyses = GREEN_LEAF_LAB_ANALYSES
    coa_parameters = GREEN_LEAF_LAB_COA

    # Get the lab's analyses.
    standard_analyses = list(lab_analyses.keys())
    analysis_names = [x['name'] for x in lab_analyses.values()]

    # Read the PDF.
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]

    # Get the lab-specific CoA page areas.
    page_area = literal_eval(coa_parameters['coa_page_area'])
    distributor_area = literal_eval(coa_parameters['coa_distributor_area'])
    producer_area = literal_eval(coa_parameters['coa_producer_area'])
    sample_details_area = literal_eval(coa_parameters['coa_sample_details_area'])

    # Get lab CoA specific fields.
    # coa_fields = coa_parameters['coa_fields']
    coa_replacements = coa_parameters['coa_replacements']
    sample_details_fields = coa_parameters['coa_sample_detail_fields']
    skip_values = coa_parameters['coa_skip_values']

    # Get all distributor details.
    # Note: This could / should be more general.
    crop = front_page.within_bbox(distributor_area)
    details = crop.extract_text().split('\n')
    address = details[2]
    parts = address.split(',')
    street = parts[0]
    subparts = parts[-1].strip().split(' ')
    city = ' '.join(subparts[:-2])
    state, zipcode = subparts[-2], subparts[-1]
    address = ','.join([street, details[3]])
    obs['distributor'] = details[1]
    obs['distributor_address'] = address
    obs['distributor_street'] = street
    obs['distributor_city'] = city
    obs['distributor_state'] = state
    obs['distributor_zipcode'] = zipcode
    obs['distributor_license_number'] = details[-1]
    
    # Get all producer details.
    # Note: This could / should be more general.
    crop = front_page.within_bbox(producer_area)
    details = crop.extract_text().split('\n')
    producer = details[1]
    street = details[2]
    parts = details[3].split(',')
    city = parts[0]
    state, zipcode = tuple(parts[-1].strip().split(' ')[-2:])
    address = ','.join([street, details[3]])
    obs['producer'] = producer
    obs['producer_address'] = address
    obs['producer_street'] = street
    obs['producer_city'] = city
    obs['producer_state'] = state
    obs['producer_zipcode'] = zipcode
    obs['producer_license_number'] = details[-1]

    # Optional: Get the image data.
    # image_index = coa_parameters['coa_image_index']
    # obs['image_data'] = parser.get_pdf_image_data(report.pages[0], image_index)
    obs['images'] = []

    # Get the sample details.
    crop = front_page.within_bbox(sample_details_area)
    details = crop.extract_text()
    details = re.split('\n|' + '|'.join(sample_details_fields), details)
    product_name = details[0]
    index = 0
    for i, detail in enumerate(details[1:]):
        if detail:
            field = sample_details_fields[index]
            key = parser.fields[field]
            obs[key] = detail.replace(':', '').strip()
            index += 1  

    # Get the `analyses` and `{analysis}_status`.
    analyses = []
    table = report.pages[0].extract_table()
    for rows in table:
        for row in rows:
            cells = row.split('\n')
            for cell in cells:
                parts = cell.split(':')
                key = snake_case(parts[0])
                try:
                    value = strip_whitespace(parts[1])
                except IndexError:
                    continue
                field = parser.fields.get(key, key)
                obs[field] = value.lower()
                if field != 'status':
                    analysis = field.replace('_status', '')
                    analyses.append(analysis)

    # Identify `{analysis}_method`s.
    # Also find the times for all of the tests and find when all
    # tests were completed. Future work: record finish time of analyses.
    methods = []
    tested_at = []
    for page in report.pages[1:]:
        text = page.extract_text()
        parts = text.split('Analysis Method/SOP')
        for part in parts[1:]:
            method = part.split('\n')[0].replace(':', '').strip()
            method = method.split('Date/Time')[0] # <-- Hot-fix.
            methods.append(method)
        try:
            parts = text.split('Analyzed:')
            date = parts[1].split('\n')[0].strip()
            tested_at.append(pd.to_datetime(date))
        except:
            pass
    date_tested = max(tested_at).isoformat()

    # Get all of the `results` rows.
    all_rows = []
    for page in report.pages[1:]:
        crop = page.within_bbox(page_area)
        rows = parser.get_page_rows(crop)
        for row in rows:
            if row in all_rows:
                pass
            else:
                all_rows.append(row)

    # Iterate over all rows to get the `results` rows
    # seeing if row starts with an analysis or analyte.
    results = []
    current_analysis = None
    for row in all_rows:

        # Identify the analysis.
        analysis = current_analysis
        for i, name in enumerate(analysis_names):
            if name in row:
                analysis = standard_analyses[i]
                break
        if analysis != current_analysis:
            current_analysis = analysis
            continue

        # Skip detail rows.
        detail_row = False
        for skip_value in skip_values:
            if row.startswith(skip_value):
                detail_row = True
                break
        if detail_row:
            continue

        # Get the analysis details.
        analysis_details = lab_analyses[current_analysis]
        columns = analysis_details['columns']
        double_column = analysis_details.get('double_column')

        # Get the result!
        values = row
        for replacement in coa_replacements:
            values = values.replace(replacement['text'], replacement['key'])
        values = values.split(' ')
        values = [x for x in values if x]
        units = STANDARD_UNITS.get(analysis)
        result = {'analysis': analysis, 'units': units}

        # Skip the analysis title row and short detail rows.
        # Note: There is probably a better way to do this that can not
        # accidentally exclude results. This method may not be perfect.
        if snake_case(values[0]) == current_analysis or len(values) < len(columns):
            continue

        # Hot-fix for `total_terpenes`.
        # Note: There may be a better way to do this.
        if values[0] == 'Total' and values[1] == 'Terpenes':
            values = ['Total Terpenes'] + values[2:]

        # Hot-fix for `mycotoxins`.
        # Note: There may be a better way to do this.
        if values[0] == 'aflatoxin':
            values.insert(2, 20)

        # Split the row if double_column.
        # Note: This code could probably be refactored.
        if double_column and len(values) > len(columns):
            multi_part = split_list(values, int(len(values) / 2))

            # Parse the first column.
            entry = augment_analyte_result(
                parser.analytes,
                result,
                columns,
                multi_part[0],
            )
            if entry['name'] != 'mgtog' and entry['name'] != 'action':
                results.append(entry)

            # Parse the second column.
            entry = augment_analyte_result(
                parser.analytes,
                result,
                columns,
                multi_part[1],
            )
            if entry['name'] != 'mgtog' and entry['name'] != 'action':
                results.append(entry)
        
        # Parse a single column.
        else:
            entry = augment_analyte_result(
                parser.analytes,
                result,
                columns,
                values,
            )
            if entry['name'] != 'mgtog' and entry['name'] != 'action':
                results.append(entry)

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass
    
    # Finish data collection with a freshly minted sample ID.
    obs['analyses'] = analyses
    obs['date_tested'] = date_tested
    obs['methods'] = methods
    obs['product_name'] = product_name
    obs['results'] = results
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=product_name,
        salt=producer,
    )
    obs['coa_parsed_at'] = datetime.now().isoformat()

    # Optional: Lowercase `status` in each of the `results`.

    # Future work: Standardize the `product_type`.

    # Future work: Attempt to identify `strain_name` from `product_name`.

    return {**GREEN_LEAF_LAB, **obs}


if __name__ == '__main__':

    # Test parsing a Green Leaf Lab CoA
    from cannlytics.data.coas import CoADoc

    # Specify where your test CoA lives.
    DATA_DIR = '../../../tests/assets/coas'
    # coa_pdf = f'{DATA_DIR}/Raspberry Parfait.pdf' # Alternative.
    coa_pdf = f'{DATA_DIR}/Flore COA/Swami Select/MagicMelon.pdf'

    # [✓] TEST: Detect the lab / LIMS that generated the CoA.
    parser = CoADoc()
    lab = parser.identify_lims(coa_pdf)
    assert lab == 'Green Leaf Lab'

    # [✓] TEST: Parse a Green Leaf Lab CoA.
    parser = CoADoc()
    data = parse_green_leaf_lab_pdf(parser, coa_pdf)
    assert data is not None



"""
Parse Modern Canna Science COAs
Copyright (c) 2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 5/20/2023
Updated: 5/21/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Modern Canna Science Labs COA PDFs.

Data Points:

    - analyses
    - {analysis}_method
    - {analysis}_status
    - coa_urls
    - date_collected
    - date_tested
    - date_received
    - distributor
    - distributor_address
    - distributor_street
    - distributor_city
    - distributor_state
    - distributor_zipcode
    - distributor_license_number
    - images
    - lab_results_url
    - producer
    - producer_address
    - producer_street
    - producer_city
    - producer_state
    - producer_zipcode
    - producer_license_number
    - product_name
    - lab_id
    - product_type
    - batch_number
    - metrc_ids
    - metrc_lab_id
    - metrc_source_id
    - product_size
    - serving_size
    - servings_per_package
    - sample_weight
    - results
    - status
    - total_cannabinoids
    - total_thc
    - total_cbd
    - total_cbg
    - total_thcv
    - total_cbc
    - total_cbdv
    - total_terpenes
    - sample_id
    - strain_name (augmented)
    - lab
    - lab_image_url
    - lab_license_number
    - lab_address
    - lab_street
    - lab_city
    - lab_county (augmented)
    - lab_state
    - lab_zipcode
    - lab_phone
    - lab_email
    - lab_website
    - lab_latitude (augmented)
    - lab_longitude (augmented)

"""
# Standard imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber

# Internal imports.
from cannlytics.data.data import create_sample_id
from cannlytics.utils.constants import STANDARD_UNITS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)

# It is assumed that the lab has the following details.
MODERNCANNA = {
    'coa_algorithm': 'moderncanna.py',
    'coa_algorithm_entry_point': 'parse_moderncanna_coa',
    'lims': 'Modern Canna Science',
    'lab': 'Modern Canna Science',
    'lab_image_url': 'https://moderncanna.com/wp-content/uploads/2019/08/logo2019.png',
    'lab_address': 'https://moderncanna.com/',
    'lab_street': '4705 Old Rd 37',
    'lab_city': 'Lakeland',
    'lab_county': 'Polk County',
    'lab_state': 'FL',
    'lab_zipcode': '33813',
    'lab_phone': '863-608-7800',
    'lab_email': 'info@moderncanna.com',
    'lab_website': 'https://moderncanna.com/',
    'lab_latitude': 27.980038,
    'lab_longitude': -81.961967,
}


def parse_moderncanna_coa(
        parser,
        doc: Any,
        temp_path: Optional[str] = None,
        **kwargs,
    ) -> dict:
    """Parse a ModernCanna Labs COA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """

    # Get the lab's analyses.
    # lab_analyses = GREEN_LEAF_LAB_ANALYSES
    # coa_parameters = GREEN_LEAF_LAB_COA
    # standard_analyses = list(lab_analyses.keys())
    # analysis_names = [x['name'] for x in lab_analyses.values()]

    # TODO: If the `doc` is a URL, then download the PDF to `temp_path`.
    if temp_path is None:
        # FIXME: Get the user's temp_path
        temp_path = '/tmp'

    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]

    # Add optional ability to collect the image.
    # - Get the image data.
    # - Save the image to Firebase Storage
    # - Create a dynamic link for the image.

    # TODO: Get the lab details.

    # TODO: Get the sample details.

    # TODO: Get the client details.

    # TODO: Get the analyses.

    # TODO: Get the results.


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc
    from dotenv import dotenv_values

    # [ ] TEST: Identify LIMS.
    parser = CoADoc()
    doc = ''
    lims = parser.identify_lims(doc, lims={'Modern Canna Science': MODERNCANNA})
    assert lims == 'Modern Canna Science'


    # [ ] TEST: Parse a Modern Canna Science COA from a URL.
    urls = [
        'https://moderncanna.com/coa/HD10010-01.pdf',
        'https://moderncanna.com/coa/HD12005-01.pdf',
        'https://moderncanna.com/coa/HD19005-11.pdf',
    ]




"""
Parse ACS Labs COAs
Copyright (c) 2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 5/18/2023
Updated: 6/5/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse ACS Labs COA PDFs.

Data Points:

    ✓ id
    ✓ lab_id
    ✓ product_name
    ✓ product_type
    ✓ batch_number
    ✓ product_size
    - serving_size
    ✓ units_per_package
    ✓ sample_weight
    ✓ strain_name
    ✓ lab
    ✓ lab_image_url
    ✓ lab_license_number
    ✓ lab_address
    ✓ lab_street
    ✓ lab_city
    ✓ lab_county (augmented)
    ✓ lab_state
    ✓ lab_zipcode
    ✓ lab_phone
    ✓ lab_email
    ✓ lab_website
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)
    ✓ producer
    - producer_address
    - producer_street
    - producer_city
    - producer_state
    - producer_zipcode
    - producer_license_number
    ✓ distributor
    - distributor_address
    - distributor_street
    - distributor_city
    - distributor_state
    - distributor_zipcode
    - distributor_license_number
    ✓ date_collected
    ✓ date_tested
    ✓ date_received
    ✓ date_harvested
    ✓ date_packaged
    - images
    ✓ analyses
    ✓ {analysis}_status
    ✓ coa_urls
    ✓ lab_results_url
    ✓ status
    - methods
    ✓ total_products
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_terpenes
    ✓ results
        ✓ cannabinoids
        ✓ terpenes
        ✓ pesticides
        ✓ heavy_metals
        ✓ microbes
        ✓ mycotoxins
        ✓ residual_solvents
        ✓ foreign_matter
        ✓ water_activity
        ✓ moisture

"""
# Standard imports.
import base64
from datetime import datetime
import json
import io
import os
import re
import tempfile
from typing import Any, List, Optional

# External imports.
import pandas as pd
import pdfplumber
from PIL import Image
import requests

# Internal imports.
from cannlytics import firebase
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
)
from cannlytics.utils.constants import DEFAULT_HEADERS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
)


def save_image_data(image_data, image_file='image.png'):
    """Save image data to a file."""
    image_bytes = base64.b64decode(image_data)
    image_io = io.BytesIO(image_bytes)
    image = Image.open(image_io)
    image.save(image_file)


def upload_image_data(storage_ref, image_file):
    """Upload image data to Firebase Storage."""
    firebase.initialize_firebase()
    firebase.upload_file(storage_ref, image_file)


def get_rows_between_values(
        original_list,
        start: Optional[str] = '',
        stop: Optional[str] = '',
    ):
    """Get rows between two values."""
    start_index = None
    end_index = None
    for index, line in enumerate(original_list):
        if line.startswith(start) and start_index is None:
            start_index = index + 1
        elif line.startswith(stop) and start_index is not None:
            end_index = index
            break
    if start_index is not None and end_index is not None:
        return original_list[start_index:end_index]
    elif start_index is not None:
        return original_list[start_index:]
    elif end_index is not None:
        return original_list[:end_index]
    else:
        return []


def split_elements(elements: list, split_words: List[str]):
    """Split elements in a list by a list of words, extending the list."""
    cells = []
    for element in elements:
        was_split = False
        temp_cells = []
        for word_index, split_word in enumerate(split_words):
            splits = element.split(split_word)
            if len(splits) > 1:  # if the element was split
                was_split = True
                for split_index, split in enumerate(splits):
                    # if it's not the first split, add split_word back
                    if split_index != 0:
                        split = split_word + split
                    temp_cells.append(split.strip())
        if was_split:  # if the element was split by any split word
            cells.extend(temp_cells)
        else:
            cells.append(element.strip())
        # remove empty strings
        cells = [i for i in cells if i]
    return cells


# It is assumed that the lab has the following details.
ACS_LABS = {
    'coa_algorithm': 'acs.py',
    'coa_algorithm_entry_point': 'parse_acs_coa',
    'lims': 'ACS Labs',
    'url': 'https://portal.acslabcannabis.com',
    'lab': 'ACS Labs',
    'lab_license_number': 'CMTL-0003',
    'lab_image_url': 'https://global-uploads.webflow.com/630470e960f8722190672cb4/6305a2e849811b34bf18777d_Desktop%20Logo.svg',
    'lab_address': '721 Cortaro Dr, Sun City Center, FL 33573',
    'lab_street': '721 Cortaro Dr',
    'lab_city': 'Sun City Center',
    'lab_county': 'Hillsborough County',
    'lab_state': 'FL',
    'lab_zipcode': '33573',
    'lab_phone': '813-634-4529',
    'lab_email': 'info@acslabcannabis.com',
    'lab_website': 'https://www.acslabcannabis.com/',
    'lab_latitude': 27.713506,
    'lab_longitude': -82.371029,
}

# It is assumed that the COA has the following details.
ACS_LABS_COA = {
    'analyses': {
        'Potency': 'cannabinoids',
        'Terpenes': 'terpenes',
        'Pesticides': 'pesticides',
        'Heavy Metals': 'heavy_metals',
        'Pathogenic': 'microbes',
        'Microbiology (qPCR)': 'microbes',
        'Microbiology Petrifilm': 'microbes',
        'Mycotoxins': 'mycotoxins',
        'Residual Solvents': 'residuals_solvents',
        'Filth and Foreign': 'foreign_matter',
        'Total Contaminant Load': 'foreign_matter',
        'Water Activity': 'water_activity',
        'Moisture': 'moisture',
    },
    'fields': {
        'Batch Date': 'date_packaged',
        'Completion Date': 'date_tested',
        'Cultivation Facility': 'producer',
        'Cultivars': 'strain_name',
        'Initial Gross Weight': 'sample_weight',
        'Lab Batch Date': 'date_received',
        'Cultivation Date': 'date_harvested',
        # FIXME: This is getting "Cultivation Date" mixed in.
        'Lot ID': 'batch_number',
        'Net Weight per Unit': 'product_size',
        'Number of Units': 'units_per_package',
        'Order #': 'lab_id',
        'Order Date': 'date_received',
        'Production Date': 'date_harvested',
        'Production Facility': 'distributor',
        'Sample #': 'id',
        'Sampling Date': 'date_collected',
        # 'Sampling Method': 'method_sampling',
        'Seed to Sale #': 'traceability_id',
        'Total Number of Final Products': 'total_products',
        'FL License #': 'lab_license_number',
        'Test Reg State': 'producer_state',
    },
}


def parse_acs_coa(
        parser,
        doc: Any,
        temp_path: Optional[str] = None,
        coa_parameters: Optional[dict] = ACS_LABS_COA,
        session: Optional[Any] = None,
        headers: Optional[dict] = DEFAULT_HEADERS,
        **kwargs,
    ) -> dict:
    """Parse a ACS Labs COA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """

    # Get the lab's parameters.
    obs = {}
    analyses, methods, results = [], [], []
    lab_analyses = coa_parameters['analyses']
    lab_fields = coa_parameters['fields']

    # If the `doc` is a URL, then download the PDF to the `temp_path`.
    # Then use the path of the downloaded PDF as the doc.
    if isinstance(doc, str):
        if doc.startswith('https'):
            if temp_path is None: temp_path = tempfile.gettempdir()
            if not os.path.exists(temp_path): os.makedirs(temp_path)
            try:
                filename = doc.split('/')[-1].split('?')[0] + '.pdf'
            except:
                filename = 'coa.pdf'
            coa_pdf = os.path.join(temp_path, filename)
            if session is not None:
                response = session.get(doc)
            else:
                response = requests.get(doc, headers=headers)
            with open(coa_pdf, 'wb') as pdf:
                pdf.write(response.content)
            report = pdfplumber.open(coa_pdf)
            obs['coa_pdf'] = filename
        else:
            report = pdfplumber.open(doc)
            obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the front page.
    page = report.pages[0]
    page_text = page.extract_text().replace(r'\u0000', '')
    lines = page_text.split('\n')
    left = page.within_bbox((0, 0, page.width * 0.5, page.height))
    right = page.within_bbox((page.width * 0.5, 0, page.width, page.height))
    rows = left.extract_text().split('\n') + right.extract_text().split('\n')

    # Get the lab results URL from the QR code.
    coa_url = parser.find_pdf_qr_code_url(page)
    if coa_url is None and doc.startswith('https'):
        coa_url = doc
    obs['lab_results_url'] = coa_url

    # Format the `coa_urls`.
    if coa_url is not None:
        filename = coa_url.split('/')[-1].split('?')[0] + '.pdf'
        obs['coa_urls'] = json.dumps([{'url': coa_url, 'filename': filename}])

    # Get the product name.
    obs['product_name'] = lines[0]

    # Get the product type.
    top_corner = page.within_bbox((page.width * 0.25, 0, page.width, page.height * 0.25))
    top_lines = top_corner.extract_text().replace('\x00', 'fi').split('\n')
    top_lines = get_rows_between_values(
        top_lines,
        start='Sample Matrix',
        stop='Certificate of Analysis'
    )
    obs['product_type'] = ' '.join(top_lines)

    # Get sample detail fields.    
    names = list(lab_fields.keys())
    for name, key in lab_fields.items():
        if name.endswith('Date'):
            values = re.findall(rf"{name}:\s+([0-9]{{4}}-[0-9]{{2}}-[0-9]{{2}})", page_text)
        else:
            values = re.findall(rf"{name}:\s+(.*)", page_text)
            if not values:
                values = re.findall(rf"{name}\s+(.*)", page_text)
                if not values:
                    values = re.findall(rf"{name}:+(.*)", page_text)
        if values:
            for n in names:
                if n != name and n in values[0]:
                    values = values[0].split(n)
            obs[key] = values[0]

    # TODO: Augment producer / distributor data.
    # - producer_address
    # - producer_street
    # - producer_city
    # - producer_state
    # - producer_zipcode
    # - producer_license_number (augment)

    # Get the analyses and statuses.
    tests = 'Potency' + page_text.split('Potency')[1].split('Product Image')[0]
    tests = tests.replace('Not Tested', 'NT')
    for k, v in lab_analyses.items():
        tests = tests.replace('\x00', 'fi').replace(k, v)
    tests = [x.split(' ') for x in tests.split('\n') if x]

    # Determine status for each analysis.
    overall_status = 'Pass'
    for index, sublist in enumerate(tests):
        if index % 2 == 0:
            analyses.extend(sublist)
        else:
            test_types = analyses[-len(sublist):]
            for k, test_types in enumerate(test_types):
                status = sublist[k].replace('Passed', 'Pass').replace('Failed', 'Fail')
                obs[f'{test_types}_status'] = status
                if 'fail' in status.lower():
                    overall_status = 'Fail'
    obs['status'] = overall_status
    analyses = list(set(analyses))

    # Get total cannabinoids.
    crop = page.within_bbox((0.5 * page.width, 0.33 * page.height, page.width, page.height * 0.8))
    cells = crop.extract_text().split('\n')
    for i, cell in enumerate(cells):

        # Get total THC and CBD.
        if cell.startswith('Total THC') or cell.startswith('Total Active THC'):
            line = cells[i + 1]
            line = line.replace('Not Detected', 'ND') \
                .replace('None Detected', 'ND') \
                .replace('-', 'ND')
            values = line.split(' ')
            obs['total_thc'] = convert_to_numeric(values[0], strip=True)
            obs['total_cbd'] = convert_to_numeric(values[-2], strip=True)
        
        # Get total cannabinoids.
        if cell.startswith('Other Cannabinoids'):
            line = cells[i + 1]
            line = line.replace('Not Detected', 'ND') \
                .replace('None Detected', 'ND') \
                .replace('-', 'ND')
            values = line.split(' ')
            obs['total_cannabinoids'] = convert_to_numeric(values[-2], strip=True)

        # Get total terpenes.
        if cell.startswith('Total Terpenes'):
            value = convert_to_numeric(cell.split(' ')[-1], strip=True)
            obs['total_terpenes'] = value
            break

    # Get cannabinoids.
    crop = page.within_bbox((0, 0.4 * page.height, page.width * 0.5, page.height * 0.8))
    cells = crop.extract_text().split('\n')
    if 'Potency' in page_text:
        stop = 'Sample Prepared By' if 'Sample Prepared By' in page_text else 'Total'
        elements = get_rows_between_values(
            rows,
            start='Analyte',
            stop=stop,
        )
        for line in elements[1:]:
            first_value = find_first_value(line)
            name = line[:first_value].strip()
            key = parser.analytes.get(snake_case(name), snake_case(name))
            values = line[first_value:].strip().split(' ')
            results.append({
                'analysis': 'cannabinoids',
                'key': key,
                'lod': convert_to_numeric(values[1]),
                'loq': convert_to_numeric(values[2]),
                'name': name,
                'units': 'percent',
                'value': convert_to_numeric(values[-1]),
                'mg_g': convert_to_numeric(values[-2]),
            })

    # Get heavy metals (flower).
    if 'Arsenic' in page.extract_text():
        elements = get_rows_between_values(
            cells,
            start='Analyte LOD',
            stop='Prep. By',
        )
        elements = split_elements(elements, ['Lead', 'Mercury'])
        for line in elements[1:]:
            first_value = find_first_value(line, [' \d+', 'ND', '<', ' .'])
            name = line[:first_value].strip()
            key = parser.analytes.get(snake_case(name), snake_case(name))
            values = line[first_value:].strip().split(' ')
            results.append({
                'analysis': 'heavy_metals',
                'key': key,
                'lod': convert_to_numeric(values[1]),
                'loq': convert_to_numeric(values[2]),
                'limit': convert_to_numeric(values[3]),
                'name': name,
                'units': 'ppb',
                'value': convert_to_numeric(values[-1]),
            })

    # Get terpene summary in case there are no terpene results later.
    terpenes_collected = []
    if 'Patient COA' in page_text:
        crop = page.within_bbox((page.width * 0.5, 0.4 * page.height, page.width, page.height * 0.8))
        cells = crop.extract_text().split('\n')
        elements = get_rows_between_values(
            cells,
            start='Analyte',
            stop='Total Terpenes',
        )
        for line in elements:
            first_value = find_first_value(line)
            name = line[:first_value].strip()
            key = parser.analytes.get(snake_case(name), snake_case(name))
            values = line[first_value:].strip().split(' ')
            results.append({
                'analysis': 'terpenes',
                'key': key,
                'name': name,
                'units': 'percent',
                'value': convert_to_numeric(values[-1], strip=True),
                'mg_g': convert_to_numeric(values[0], strip=True),
            })
            terpenes_collected.append(key)

    # Get results.
    for page in report.pages[1:]:
        
        # Get page rows from the double column layout.
        page_text = page.extract_text().replace(r'\u0000', '')
        left = page.within_bbox((0, 0, page.width * 0.5, page.height))
        right = page.within_bbox((page.width * 0.5, 0, page.width, page.height))
        rows = left.extract_text().replace(r'\u0000', '').split('\n')
        rows += right.extract_text().replace(r'\u0000', '').split('\n')

        # Handle wide first column for mycotoxins.
        if 'Mycotoxins' in page_text and 'Prep. By' in page_text:
            left = page.within_bbox((0, 0, page.width * 0.5725, page.height))
            right = page.within_bbox((page.width * 0.5725, 0, page.width, page.height))
            rows = left.extract_text().split('\n') + right.extract_text().split('\n')

        # Get residual solvents (concentrates).
        # FIXME: Test on recent COAs.
        if 'Residual Solvents' in page_text:
            q3 = page.within_bbox((page.width * .5, 0, page.width * .72, page.height))
            q4 = page.within_bbox((page.width * .71, 0, page.width, page.height))
            texts = '\n'.join([q3.extract_text(), q4.extract_text()])
            cells = texts.split('\n')
            elements = get_rows_between_values(
                cells,
                start='Analyte',
                stop='Lab Batch #',
            )
            set1 = get_rows_between_values(
                elements,
                stop='Sample Prepared By',
            )
            set2 = get_rows_between_values(
                elements,
                start='Analyte (1:n)',
                stop='Sample Analyzed By',
            )
            table = [x for x in set1 + set2 if len(x.split(' ')) > 1]
            for line in table:
                if line.startswith('1,'):
                    line = line.replace('- ', '-Dichloroethane ')
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                results.append({
                    'analysis': 'residual_solvents',
                    'key': key,
                    'name': name,
                    'units': 'ppm',
                    'loq': convert_to_numeric(values[-3]),
                    'limit': convert_to_numeric(values[-2]),
                    'value': convert_to_numeric(values[-1]),
                })

        # Get microbes.
        if 'Pathogenic' in page_text:
            first_stop = 'Reviewed By' if 'Reviewed By' in page_text else 'Batch Reviewed By'
            stop = 'Prep. By' if 'Prep. By' in page_text else 'Sample Prepared By'
            elements = get_rows_between_values(
                get_rows_between_values(rows, start='Pathogenic', stop=first_stop),
                start='Analyte',
                stop=stop,
            )
            elements = [x.replace('\x00', 'fl') for x in elements if len(x.split(' ')) > 1]
            elements = split_elements(elements, ['Aspergillus', 'Salmonella', 'STEC'])
            elements = [x.replace('Absence in 1g', 'ND') \
                .replace('Absence in 1 g', 'ND') \
                .replace('Absence', 'ND') \
                .replace('in 1 g', 'ND') for x in elements]
            for line in elements[1:]:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                if len(values) > 2:
                    continue
                results.append({
                    'analysis': 'microbes',
                    'key': key,
                    'name': name,
                    'units': 'cfu/g',
                    'value': convert_to_numeric(values[-1]),
                    'limit': convert_to_numeric(values[0]),
                })

        # Get moisture content (flower).
        if 'Moisture' in page_text:
            for row in rows:
                if row.startswith('Moisture 15'):
                    first_value = find_first_value(row)
                    name = row[:first_value].strip()
                    key = parser.analytes.get(snake_case(name), snake_case(name))
                    values = row[first_value:].strip().split(' ')
                    results.append({
                        'analysis': 'moisture',
                        'key': key,
                        'name': name,
                        'units': 'percent',
                        'value': convert_to_numeric(values[-1]),
                        'limit': convert_to_numeric(values[0]),
                    })
        
        # Get terpenes.
        if 'Terpenes' in page_text:
            elements = []
            if 'Sample Prepared By' in page_text:
                elements = get_rows_between_values(
                    rows,
                    start='Analyte Dilution',
                    stop='Sample Prepared By',
                )
            elif 'Prep. By' in page_text:
                set1 = get_rows_between_values(
                    rows,
                    start='Analyte',
                    stop='Prep. By',
                )
                set2 = get_rows_between_values(
                    get_rows_between_values(rows, start='Analyte'),
                    start='Analyte',
                    stop='Analyzed By',
                )
                elements = [x for x in set1 + set2 if len(x.split(' ')) > 1]
            for line in elements[1:]:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                if key in terpenes_collected: # Skip any terpenes collected in the summary.
                    continue
                values = line[first_value:].strip().split(' ')
                results.append({
                    'analysis': 'terpenes',
                    'key': key,
                    'lod': convert_to_numeric(values[1]),
                    'name': name,
                    'units': 'percent',
                    'value': convert_to_numeric(values[-1]),
                })

        # Get mycotoxins.
        if 'Mycotoxins' in page_text:
            stop = 'Prep. By' if 'Prep. By' in page_text else 'Sample Prepared By'
            elements = get_rows_between_values(
                rows,
                start='Mycotoxins',
                stop='Batch Reviewed',
            )
            elements = get_rows_between_values(
                elements,
                start='Analyte',
                stop=stop,
            )
            elements = [x.replace('\x00', 'fl') for x in elements]
            cells = split_elements(elements, ['Aflatoxin', 'Ochratoxin'])
            for line in cells[1:]:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                if len(values) > 4:
                    continue
                results.append({
                    'analysis': 'mycotoxins',
                    'key': key,
                    'lod': convert_to_numeric(values[1]),
                    'name': name,
                    'units': 'ppb',
                    'value': convert_to_numeric(values[-1]),
                    'limit': convert_to_numeric(values[-2]),
                })

        # Get heavy metals.
        if 'Arsenic' in page_text:
            stop = 'Prep. By' if 'Prep. By' in page_text else 'Sample Prepared By'
            elements = get_rows_between_values(
                rows,
                start='Heavy Metals',
                stop='Batch Reviewed',
            )
            elements = get_rows_between_values(
                elements,
                start='Analyte',
                stop=stop,
            )
            cells = split_elements(elements, ['Lead', 'Mercury'])
            for line in cells[1:]:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                results.append({
                    'analysis': 'heavy_metals',
                    'key': key,
                    'lod': convert_to_numeric(values[0]),
                    'name': name,
                    'units': 'ppb',
                    'value': convert_to_numeric(values[-1]),
                    'limit': convert_to_numeric(values[1]),
                })

        # Get pesticides.
        if 'Pesticides' in page_text:
            if 'Sample Prepared By' in page_text:
                q1 = page.within_bbox((0, 0, page.width * 0.27, page.height))
                q2 = page.within_bbox((page.width * 0.27, 0, page.width * 0.5, page.height))
                texts = '\n'.join([q1.extract_text(), q2.extract_text()])
                texts = texts.replace('\x00', 'fl')
                cells = texts.split('\n')
                elements = get_rows_between_values(
                    cells,
                    start='Analyte',
                    stop='Batch Reviewed By',
                )
                set1 = get_rows_between_values(
                    elements,
                    stop='Sample Prepared By',
                )
                set2 = get_rows_between_values(
                    elements,
                    start='Analyte',
                    stop='Sample Analyzed By',
                )
                table = [x for x in set1 + set2 if len(x.split(' ')) > 1]
            else:
                set1 = get_rows_between_values(
                    rows,
                    start='Analyte',
                    stop='Prep. By',
                )
                set2 = get_rows_between_values(
                    get_rows_between_values(rows, start='Prep. By', stop='Lab Batch #'),
                    start='Analyte',
                    stop='Analyzed By',
                )
                elements = [x.replace('\x00', 'fl') for x in set1 + set2 if len(x.split(' ')) > 1]
                table = [x for x in elements if '(ppb)' not in x]
            for line in table:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                if len(values) < 3:
                    continue
                results.append({
                    'analysis': 'pesticides',
                    'key': key,
                    'name': name,
                    'units': 'ppb',
                    'loq': convert_to_numeric(values[1]),
                    'limit': convert_to_numeric(values[2]),
                    'value': convert_to_numeric(values[-1]),
                })

        # Get foreign matter.
        if 'Filth and Foreign Material' in page_text:
            stop = 'Prep. By' if 'Prep. By' in page_text else 'Sample Prepared By'
            elements = get_rows_between_values(
                rows,
                start='Filth and Foreign Material',
                stop='Batch Reviewed',
            )
            elements = get_rows_between_values(
                elements,
                start='Analyte',
                stop=stop,
            )
            split_words = ['Weight']
            cells = split_elements(elements, split_words)
            for line in cells[1:]:
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                results.append({
                    'analysis': 'foreign_matter',
                    'key': key,
                    'name': name,
                    'units': 'percent',
                    'value': convert_to_numeric(values[-1]),
                    'limit': convert_to_numeric(values[0]),
                })

        # Get water activity.
        if 'Water Activity' in page_text:
            for row in rows:
                if row.startswith('Water Activity 0.65'):
                    first_value = find_first_value(row)
                    name = row[:first_value].strip()
                    key = parser.analytes.get(snake_case(name), snake_case(name))
                    values = row[first_value:].strip().split(' ')
                    results.append({
                        'analysis': 'water_activity',
                        'key': key,
                        'name': name,
                        'units': 'aw',
                        'value': convert_to_numeric(values[-1]),
                        'limit': convert_to_numeric(values[0]),
                    })

        # Get total yeast and mold.
        if 'Total Yeast and Mold' in page_text:
            for row in rows:
                if row.startswith('Total Yeast/Mold'):
                    first_value = find_first_value(row)
                    name = row[:first_value].strip()
                    key = parser.analytes.get(snake_case(name), snake_case(name))
                    values = row[first_value:].strip().split(' ')
                    status = values[-1].replace('Passed', 'Pass').replace('Failed', 'Fail')
                    result = {
                        'analysis': 'microbe',
                        'key': key,
                        'name': name,
                        'units': 'CFU/g',
                        'value': convert_to_numeric(status),
                        'status': status,
                        'limit': convert_to_numeric(values[0]),
                    }
                    if len(values) == 3:
                        result['value'] = convert_to_numeric(values[-2])
                    results.append(result)

    # Optional: Get the image data.

    # Close the report.
    report.close()

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs = {**ACS_LABS, **obs}
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['methods'] = json.dumps(methods)
    obs['results'] = json.dumps(results)
    obs['results_hash'] = create_hash(obs['results'])
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


# === Tests ===
if __name__ == '__main__':

    # Initialize CoADoc.
    from cannlytics.data.coas import CoADoc
    from time import sleep
    parser = CoADoc()

    # [✓] TEST: Identify LIMS from a COA URL.
    url = 'https://portal.acslabcannabis.com/qr-coa-view?salt=QUFFQzc0OS0wMTA1MjMtU0dMQzEyLVIzNS0wMjIxMjAyMw=='
    lims = parser.identify_lims(url, lims={'ACS Labs': ACS_LABS})
    assert lims == 'ACS Labs'
    print('Identified LIMS:', lims)

    # [✓] TEST: Identify LIMS from a COA PDF.
    parser = CoADoc()
    doc = '../../../../tests/assets/coas/acs/AAEC749-010523-SGLC12-R35-02212023-COA_EN.pdf'
    lims = parser.identify_lims(doc, lims={'ACS Labs': ACS_LABS})
    assert lims == 'ACS Labs'
    print('Identified LIMS:', lims)

    # [✓] TEST: Parse partial COA.
    doc = '../../../../tests/assets/coas/acs/AAEC749-010523-SGLC12-R35-02212023-COA_EN.pdf'
    data = parse_acs_coa(parser, doc)
    assert data is not None
    print('Parsed:', data)

    # [✓] TEST: Parse full panel COA for a concentrate.
    doc = '../../../../tests/assets/coas/acs/27675_0002407047.pdf'
    data = parse_acs_coa(parser, doc)
    assert data is not None
    print('Parsed:', data)

    # [✓] TEST: Parse a full panel COA for a flower.
    doc = '../../../../tests/assets/coas/acs/49448_0004136268.pdf'
    data = parse_acs_coa(parser, doc)
    assert data is not None
    print('Parsed:', data)

    # [✓] TEST: Parse partial COA from a URL.
    urls = [
        'https://portal.acslabcannabis.com/qr-coa-view?salt=QUFEVjIxMC0xMDI1MjItREJGSy1SMzUtMTIxMTIwMjI=',
        'https://www.trulieve.com/files/lab-results/18362_0003059411.pdf',
    ]
    for url in urls:
        data = parse_acs_coa(parser, url)
        assert data is not None
        print('Parsed:', data)
        sleep(3)

    # [✓] TEST: Parse a full panel COA from a URL.
    urls = [
        'https://www.trulieve.com/files/lab-results/27675_0002407047.pdf',
    ]
    for url in urls:
        data = parse_acs_coa(parser, url)
        assert data is not None
        print('Parsed:', data)
        sleep(3)

"""
# EXAMPLE: Parse a folder of ACS labs COAs.

from datetime import datetime
import os
import pandas as pd
from cannlytics.data.coas import CoADoc

# Initialize CoADoc.
parser = CoADoc()

# Specify where your ACS Labs COAs live.
all_data = []
data_dir = 'D://data/florida/lab_results/.datasets/pdfs/acs'
coa_pdfs = os.listdir(data_dir)
for coa_pdf in coa_pdfs:
    filename = os.path.join(data_dir, coa_pdf)
    try:
        data = parser.parse(filename)
        all_data.extend(data)
        print('Parsed:', filename)
    except:
        print('Failed to parse:', filename)

# Save the data.
date = datetime.now().isoformat()[:19].replace(':', '-')
outfile = f'D://data/florida/lab_results/.datasets/acs-lab-results-{date}.xlsx'
df = pd.DataFrame(all_data)
df.replace(r'\\u0000', '', regex=True, inplace=True)
parser.save(df, outfile)
"""
